# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import os.path as op
import shutil

import numpy as np
import nibabel as nb
import networkx as nx

from ... import logging
from ..base import (
    BaseInterface,
    LibraryBaseInterface,
    BaseInterfaceInputSpec,
    traits,
    File,
    TraitedSpec,
    Directory,
    isdefined,
)
from .base import have_cmp

iflogger = logging.getLogger("nipype.interface")


def create_annot_label(subject_id, subjects_dir, fs_dir, parcellation_name):
    import cmp
    from cmp.util import runCmd

    iflogger.info("Create the cortical labels necessary for our ROIs")
    iflogger.info("=================================================")
    fs_label_dir = op.join(op.join(subjects_dir, subject_id), "label")
    output_dir = op.abspath(op.curdir)
    paths = []
    cmp_config = cmp.configuration.PipelineConfiguration()
    cmp_config.parcellation_scheme = "Lausanne2008"
    for hemi in ["lh", "rh"]:
        spath = (
            cmp_config._get_lausanne_parcellation("Lausanne2008")[parcellation_name][
                "fs_label_subdir_name"
            ]
            % hemi
        )
        paths.append(spath)
    for p in paths:
        try:
            os.makedirs(op.join(".", p))
        except:
            pass
    if "33" in parcellation_name:
        comp = [
            (
                "rh",
                "myatlas_36_rh.gcs",
                "rh.myaparc_36.annot",
                "regenerated_rh_36",
                "myaparc_36",
            ),
            (
                "rh",
                "myatlas_60_rh.gcs",
                "rh.myaparc_60.annot",
                "regenerated_rh_60",
                "myaparc_60",
            ),
            (
                "lh",
                "myatlas_36_lh.gcs",
                "lh.myaparc_36.annot",
                "regenerated_lh_36",
                "myaparc_36",
            ),
            (
                "lh",
                "myatlas_60_lh.gcs",
                "lh.myaparc_60.annot",
                "regenerated_lh_60",
                "myaparc_60",
            ),
        ]
    elif "60" in parcellation_name:
        comp = [
            (
                "rh",
                "myatlas_60_rh.gcs",
                "rh.myaparc_60.annot",
                "regenerated_rh_60",
                "myaparc_60",
            ),
            (
                "lh",
                "myatlas_60_lh.gcs",
                "lh.myaparc_60.annot",
                "regenerated_lh_60",
                "myaparc_60",
            ),
        ]
    elif "125" in parcellation_name:
        comp = [
            (
                "rh",
                "myatlas_125_rh.gcs",
                "rh.myaparc_125.annot",
                "regenerated_rh_125",
                "myaparc_125",
            ),
            (
                "rh",
                "myatlas_60_rh.gcs",
                "rh.myaparc_60.annot",
                "regenerated_rh_60",
                "myaparc_60",
            ),
            (
                "lh",
                "myatlas_125_lh.gcs",
                "lh.myaparc_125.annot",
                "regenerated_lh_125",
                "myaparc_125",
            ),
            (
                "lh",
                "myatlas_60_lh.gcs",
                "lh.myaparc_60.annot",
                "regenerated_lh_60",
                "myaparc_60",
            ),
        ]
    elif "250" in parcellation_name:
        comp = [
            (
                "rh",
                "myatlas_250_rh.gcs",
                "rh.myaparc_250.annot",
                "regenerated_rh_250",
                "myaparc_250",
            ),
            (
                "rh",
                "myatlas_60_rh.gcs",
                "rh.myaparc_60.annot",
                "regenerated_rh_60",
                "myaparc_60",
            ),
            (
                "lh",
                "myatlas_250_lh.gcs",
                "lh.myaparc_250.annot",
                "regenerated_lh_250",
                "myaparc_250",
            ),
            (
                "lh",
                "myatlas_60_lh.gcs",
                "lh.myaparc_60.annot",
                "regenerated_lh_60",
                "myaparc_60",
            ),
        ]
    else:
        comp = [
            (
                "rh",
                "myatlas_36_rh.gcs",
                "rh.myaparc_36.annot",
                "regenerated_rh_36",
                "myaparc_36",
            ),
            (
                "rh",
                "myatlasP1_16_rh.gcs",
                "rh.myaparcP1_16.annot",
                "regenerated_rh_500",
                "myaparcP1_16",
            ),
            (
                "rh",
                "myatlasP17_28_rh.gcs",
                "rh.myaparcP17_28.annot",
                "regenerated_rh_500",
                "myaparcP17_28",
            ),
            (
                "rh",
                "myatlasP29_36_rh.gcs",
                "rh.myaparcP29_36.annot",
                "regenerated_rh_500",
                "myaparcP29_36",
            ),
            (
                "rh",
                "myatlas_60_rh.gcs",
                "rh.myaparc_60.annot",
                "regenerated_rh_60",
                "myaparc_60",
            ),
            (
                "rh",
                "myatlas_125_rh.gcs",
                "rh.myaparc_125.annot",
                "regenerated_rh_125",
                "myaparc_125",
            ),
            (
                "rh",
                "myatlas_250_rh.gcs",
                "rh.myaparc_250.annot",
                "regenerated_rh_250",
                "myaparc_250",
            ),
            (
                "lh",
                "myatlas_36_lh.gcs",
                "lh.myaparc_36.annot",
                "regenerated_lh_36",
                "myaparc_36",
            ),
            (
                "lh",
                "myatlasP1_16_lh.gcs",
                "lh.myaparcP1_16.annot",
                "regenerated_lh_500",
                "myaparcP1_16",
            ),
            (
                "lh",
                "myatlasP17_28_lh.gcs",
                "lh.myaparcP17_28.annot",
                "regenerated_lh_500",
                "myaparcP17_28",
            ),
            (
                "lh",
                "myatlasP29_36_lh.gcs",
                "lh.myaparcP29_36.annot",
                "regenerated_lh_500",
                "myaparcP29_36",
            ),
            (
                "lh",
                "myatlas_60_lh.gcs",
                "lh.myaparc_60.annot",
                "regenerated_lh_60",
                "myaparc_60",
            ),
            (
                "lh",
                "myatlas_125_lh.gcs",
                "lh.myaparc_125.annot",
                "regenerated_lh_125",
                "myaparc_125",
            ),
            (
                "lh",
                "myatlas_250_lh.gcs",
                "lh.myaparc_250.annot",
                "regenerated_lh_250",
                "myaparc_250",
            ),
        ]

    log = cmp_config.get_logger()

    for out in comp:
        mris_cmd = 'mris_ca_label %s %s "%s/surf/%s.sphere.reg" "%s" "%s" ' % (
            subject_id,
            out[0],
            op.join(subjects_dir, subject_id),
            out[0],
            cmp_config.get_lausanne_atlas(out[1]),
            op.join(fs_label_dir, out[2]),
        )
        runCmd(mris_cmd, log)
        iflogger.info("-----------")

        annot = '--annotation "%s"' % out[4]

        mri_an_cmd = 'mri_annotation2label --subject %s --hemi %s --outdir "%s" %s' % (
            subject_id,
            out[0],
            op.join(output_dir, out[3]),
            annot,
        )
        iflogger.info(mri_an_cmd)
        runCmd(mri_an_cmd, log)
        iflogger.info("-----------")
        iflogger.info(os.environ["SUBJECTS_DIR"])
        # extract cc and unknown to add to tractography mask, we do not want this as a region of interest
        # in FS 5.0, unknown and corpuscallosum are not available for the 35 scale (why?),
        # but for the other scales only, take the ones from _60
        rhun = op.join(output_dir, "rh.unknown.label")
        lhun = op.join(output_dir, "lh.unknown.label")
        rhco = op.join(output_dir, "rh.corpuscallosum.label")
        lhco = op.join(output_dir, "lh.corpuscallosum.label")
    shutil.copy(op.join(output_dir, "regenerated_rh_60", "rh.unknown.label"), rhun)
    shutil.copy(op.join(output_dir, "regenerated_lh_60", "lh.unknown.label"), lhun)
    shutil.copy(
        op.join(output_dir, "regenerated_rh_60", "rh.corpuscallosum.label"), rhco
    )
    shutil.copy(
        op.join(output_dir, "regenerated_lh_60", "lh.corpuscallosum.label"), lhco
    )

    mri_cmd = """mri_label2vol --label "%s" --label "%s" --label "%s" --label "%s" --temp "%s" --o  "%s" --identity """ % (
        rhun,
        lhun,
        rhco,
        lhco,
        op.join(op.join(subjects_dir, subject_id), "mri", "orig.mgz"),
        op.join(fs_label_dir, "cc_unknown.nii.gz"),
    )
    runCmd(mri_cmd, log)
    runCmd("mris_volmask %s" % subject_id, log)
    mri_cmd = 'mri_convert -i "%s/mri/ribbon.mgz" -o "%s/mri/ribbon.nii.gz"' % (
        op.join(subjects_dir, subject_id),
        op.join(subjects_dir, subject_id),
    )
    runCmd(mri_cmd, log)
    mri_cmd = 'mri_convert -i "%s/mri/aseg.mgz" -o "%s/mri/aseg.nii.gz"' % (
        op.join(subjects_dir, subject_id),
        op.join(subjects_dir, subject_id),
    )
    runCmd(mri_cmd, log)

    iflogger.info("[ DONE ]")


def create_roi(subject_id, subjects_dir, fs_dir, parcellation_name, dilation):
    """Creates the ROI_%s.nii.gz files using the given parcellation information
    from networks. Iteratively create volume."""
    import cmp
    from cmp.util import runCmd

    iflogger.info("Create the ROIs:")
    output_dir = op.abspath(op.curdir)
    fs_dir = op.join(subjects_dir, subject_id)
    cmp_config = cmp.configuration.PipelineConfiguration()
    cmp_config.parcellation_scheme = "Lausanne2008"
    log = cmp_config.get_logger()
    parval = cmp_config._get_lausanne_parcellation("Lausanne2008")[parcellation_name]
    pgpath = parval["node_information_graphml"]
    aseg = nb.load(op.join(fs_dir, "mri", "aseg.nii.gz"))
    asegd = np.asanyarray(aseg.dataobj)

    # identify cortical voxels, right (3) and left (42) hemispheres
    idxr = np.where(asegd == 3)
    idxl = np.where(asegd == 42)
    xx = np.concatenate((idxr[0], idxl[0]))
    yy = np.concatenate((idxr[1], idxl[1]))
    zz = np.concatenate((idxr[2], idxl[2]))

    # initialize variables necessary for cortical ROIs dilation
    # dimensions of the neighbourhood for rois labels assignment (choose odd dimensions!)
    shape = (25, 25, 25)
    center = np.array(shape) // 2
    # dist: distances from the center of the neighbourhood
    dist = np.zeros(shape, dtype="float32")
    for x in range(shape[0]):
        for y in range(shape[1]):
            for z in range(shape[2]):
                distxyz = center - [x, y, z]
                dist[x, y, z] = np.sqrt(np.sum(np.multiply(distxyz, distxyz)))

    iflogger.info("Working on parcellation: ")
    iflogger.info(
        cmp_config._get_lausanne_parcellation("Lausanne2008")[parcellation_name]
    )
    iflogger.info("========================")
    pg = nx.read_graphml(pgpath)
    # each node represents a brain region
    # create a big 256^3 volume for storage of all ROIs
    rois = np.zeros((256, 256, 256), dtype=np.int16)

    count = 0
    for brk, brv in pg.nodes(data=True):
        count = count + 1
        iflogger.info(brv)
        iflogger.info(brk)
        if brv["dn_hemisphere"] == "left":
            hemi = "lh"
        elif brv["dn_hemisphere"] == "right":
            hemi = "rh"
        if brv["dn_region"] == "subcortical":
            iflogger.info(brv)
            iflogger.info("---------------------")
            iflogger.info("Work on brain region: %s", brv["dn_region"])
            iflogger.info("Freesurfer Name: %s", brv["dn_fsname"])
            iflogger.info("Region %s of %s", count, pg.number_of_nodes())
            iflogger.info("---------------------")
            # if it is subcortical, retrieve roi from aseg
            idx = np.where(asegd == int(brv["dn_fs_aseg_val"]))
            rois[idx] = int(brv["dn_correspondence_id"])

        elif brv["dn_region"] == "cortical":
            iflogger.info(brv)
            iflogger.info("---------------------")
            iflogger.info("Work on brain region: %s", brv["dn_region"])
            iflogger.info("Freesurfer Name: %s", brv["dn_fsname"])
            iflogger.info("Region %s of %s", count, pg.number_of_nodes())
            iflogger.info("---------------------")

            labelpath = op.join(output_dir, parval["fs_label_subdir_name"] % hemi)
            # construct .label file name

            fname = "%s.%s.label" % (hemi, brv["dn_fsname"])

            # execute fs mri_label2vol to generate volume roi from the label file
            # store it in temporary file to be overwritten for each region

            mri_cmd = 'mri_label2vol --label "%s" --temp "%s" --o "%s" --identity' % (
                op.join(labelpath, fname),
                op.join(fs_dir, "mri", "orig.mgz"),
                op.join(output_dir, "tmp.nii.gz"),
            )
            runCmd(mri_cmd, log)

            tmp = nb.load(op.join(output_dir, "tmp.nii.gz"))
            tmpd = np.asanyarray(tmp.dataobj)

            # find voxel and set them to intensityvalue in rois
            idx = np.where(tmpd == 1)
            rois[idx] = int(brv["dn_correspondence_id"])

        # store volume eg in ROI_scale33.nii.gz
        out_roi = op.abspath("ROI_%s.nii.gz" % parcellation_name)

        # update the header
        hdr = aseg.header
        hdr2 = hdr.copy()
        hdr2.set_data_dtype(np.uint16)

        log.info("Save output image to %s" % out_roi)
        img = nb.Nifti1Image(rois, aseg.affine, hdr2)
        nb.save(img, out_roi)

    iflogger.info("[ DONE ]")
    # dilate cortical regions
    if dilation is True:
        iflogger.info("Dilating cortical regions...")
        # loop throughout all the voxels belonging to the aseg GM volume
        for j in range(xx.size):
            if rois[xx[j], yy[j], zz[j]] == 0:
                local = extract(rois, shape, position=(xx[j], yy[j], zz[j]), fill=0)
                mask = local.copy()
                mask[np.nonzero(local > 0)] = 1
                thisdist = np.multiply(dist, mask)
                thisdist[np.nonzero(thisdist == 0)] = np.amax(thisdist)
                value = np.int_(local[np.nonzero(thisdist == np.amin(thisdist))])
                if value.size > 1:
                    counts = np.bincount(value)
                    value = np.argmax(counts)
                rois[xx[j], yy[j], zz[j]] = value

        # store volume eg in ROIv_scale33.nii.gz
        out_roi = op.abspath("ROIv_%s.nii.gz" % parcellation_name)
        iflogger.info("Save output image to %s", out_roi)
        img = nb.Nifti1Image(rois, aseg.affine, hdr2)
        nb.save(img, out_roi)

        iflogger.info("[ DONE ]")


def create_wm_mask(subject_id, subjects_dir, fs_dir, parcellation_name):
    import cmp
    import scipy.ndimage.morphology as nd

    iflogger.info("Create white matter mask")
    fs_dir = op.join(subjects_dir, subject_id)
    cmp_config = cmp.configuration.PipelineConfiguration()
    cmp_config.parcellation_scheme = "Lausanne2008"
    pgpath = cmp_config._get_lausanne_parcellation("Lausanne2008")[parcellation_name][
        "node_information_graphml"
    ]
    # load ribbon as basis for white matter mask
    fsmask = nb.load(op.join(fs_dir, "mri", "ribbon.nii.gz"))
    fsmaskd = np.asanyarray(fsmask.dataobj)

    wmmask = np.zeros(fsmaskd.shape)
    # extract right and left white matter
    idx_lh = np.where(fsmaskd == 120)
    idx_rh = np.where(fsmaskd == 20)

    wmmask[idx_lh] = 1
    wmmask[idx_rh] = 1

    # remove subcortical nuclei from white matter mask
    aseg = nb.load(op.join(fs_dir, "mri", "aseg.nii.gz"))
    asegd = np.asanyarray(aseg.dataobj)

    # need binary erosion function
    imerode = nd.binary_erosion

    # ventricle erosion
    csfA = np.zeros(asegd.shape)
    csfB = np.zeros(asegd.shape)

    # structuring elements for erosion
    se1 = np.zeros((3, 3, 5))
    se1[1, :, 2] = 1
    se1[:, 1, 2] = 1
    se1[1, 1, :] = 1
    se = np.zeros((3, 3, 3))
    se[1, :, 1] = 1
    se[:, 1, 1] = 1
    se[1, 1, :] = 1

    # lateral ventricles, thalamus proper and caudate
    # the latter two removed for better erosion, but put back afterwards
    idx = np.where(
        (asegd == 4)
        | (asegd == 43)
        | (asegd == 11)
        | (asegd == 50)
        | (asegd == 31)
        | (asegd == 63)
        | (asegd == 10)
        | (asegd == 49)
    )
    csfA[idx] = 1
    csfA = imerode(imerode(csfA, se1), se)

    # thalmus proper and cuadate are put back because they are not lateral ventricles
    idx = np.where((asegd == 11) | (asegd == 50) | (asegd == 10) | (asegd == 49))
    csfA[idx] = 0

    # REST CSF, IE 3RD AND 4TH VENTRICULE AND EXTRACEREBRAL CSF
    idx = np.where(
        (asegd == 5)
        | (asegd == 14)
        | (asegd == 15)
        | (asegd == 24)
        | (asegd == 44)
        | (asegd == 72)
        | (asegd == 75)
        | (asegd == 76)
        | (asegd == 213)
        | (asegd == 221)
    )
    # 43 ??, 4??  213?, 221?
    # more to discuss.
    for i in [5, 14, 15, 24, 44, 72, 75, 76, 213, 221]:
        idx = np.where(asegd == i)
        csfB[idx] = 1

    # do not remove the subthalamic nucleus for now from the wm mask
    # 23, 60
    # would stop the fiber going to the segmented "brainstem"

    # grey nuclei, either with or without erosion
    gr_ncl = np.zeros(asegd.shape)

    # with erosion
    for i in [10, 11, 12, 49, 50, 51]:
        idx = np.where(asegd == i)
        # temporary volume
        tmp = np.zeros(asegd.shape)
        tmp[idx] = 1
        tmp = imerode(tmp, se)
        idx = np.where(tmp == 1)
        gr_ncl[idx] = 1

    # without erosion
    for i in [13, 17, 18, 26, 52, 53, 54, 58]:
        idx = np.where(asegd == i)
        gr_ncl[idx] = 1

    # remove remaining structure, e.g. brainstem
    remaining = np.zeros(asegd.shape)
    idx = np.where(asegd == 16)
    remaining[idx] = 1

    # now remove all the structures from the white matter
    idx = np.where((csfA != 0) | (csfB != 0) | (gr_ncl != 0) | (remaining != 0))
    wmmask[idx] = 0
    iflogger.info(
        "Removing lateral ventricles and eroded grey nuclei and brainstem from white matter mask"
    )

    # ADD voxels from 'cc_unknown.nii.gz' dataset
    ccun = nb.load(op.join(fs_dir, "label", "cc_unknown.nii.gz"))
    ccund = np.asanyarray(ccun.dataobj)
    idx = np.where(ccund != 0)
    iflogger.info("Add corpus callosum and unknown to wm mask")
    wmmask[idx] = 1

    # check if we should subtract the cortical rois from this parcellation
    iflogger.info(
        "Loading ROI_%s.nii.gz to subtract cortical ROIs from white " "matter mask",
        parcellation_name,
    )
    roi = nb.load(op.join(op.curdir, "ROI_%s.nii.gz" % parcellation_name))
    roid = np.asanyarray(roi.dataobj)
    assert roid.shape[0] == wmmask.shape[0]
    pg = nx.read_graphml(pgpath)
    for brk, brv in pg.nodes(data=True):
        if brv["dn_region"] == "cortical":
            iflogger.info(
                "Subtracting region %s with intensity value %s",
                brv["dn_region"],
                brv["dn_correspondence_id"],
            )
            idx = np.where(roid == int(brv["dn_correspondence_id"]))
            wmmask[idx] = 0

    # output white matter mask. crop and move it afterwards
    wm_out = op.join(fs_dir, "mri", "fsmask_1mm.nii.gz")
    img = nb.Nifti1Image(wmmask, fsmask.affine, fsmask.header)
    iflogger.info("Save white matter mask: %s", wm_out)
    nb.save(img, wm_out)


def crop_and_move_datasets(
    subject_id, subjects_dir, fs_dir, parcellation_name, out_roi_file, dilation
):
    from cmp.util import runCmd

    fs_dir = op.join(subjects_dir, subject_id)
    cmp_config = cmp.configuration.PipelineConfiguration()
    cmp_config.parcellation_scheme = "Lausanne2008"
    log = cmp_config.get_logger()
    output_dir = op.abspath(op.curdir)

    iflogger.info("Cropping and moving datasets to %s", output_dir)
    ds = [
        (op.join(fs_dir, "mri", "aseg.nii.gz"), op.abspath("aseg.nii.gz")),
        (op.join(fs_dir, "mri", "ribbon.nii.gz"), op.abspath("ribbon.nii.gz")),
        (op.join(fs_dir, "mri", "fsmask_1mm.nii.gz"), op.abspath("fsmask_1mm.nii.gz")),
        (
            op.join(fs_dir, "label", "cc_unknown.nii.gz"),
            op.abspath("cc_unknown.nii.gz"),
        ),
    ]

    ds.append(
        (
            op.abspath("ROI_%s.nii.gz" % parcellation_name),
            op.abspath("ROI_HR_th.nii.gz"),
        )
    )
    if dilation is True:
        ds.append(
            (
                op.abspath("ROIv_%s.nii.gz" % parcellation_name),
                op.abspath("ROIv_HR_th.nii.gz"),
            )
        )
    orig = op.join(fs_dir, "mri", "orig", "001.mgz")
    for d in ds:
        iflogger.info("Processing %s:", d[0])
        if not op.exists(d[0]):
            raise Exception("File %s does not exist." % d[0])
        # reslice to original volume because the roi creation with freesurfer
        # changed to 256x256x256 resolution
        mri_cmd = 'mri_convert -rl "%s" -rt nearest "%s" -nc "%s"' % (orig, d[0], d[1])
        runCmd(mri_cmd, log)


def extract(Z, shape, position, fill):
    """Extract voxel neighbourhood

    Parameters
    ----------
    Z : array-like
        the original data
    shape : tuple
        tuple containing neighbourhood dimensions
    position : tuple
        tuple containing central point indexes
    fill : float
       value for the padding of Z

    Returns
    -------
    R : ndarray
        the neighbourhood of the specified point in Z

    """
    R = (
        np.ones(shape, dtype=Z.dtype) * fill
    )  # initialize output block to the fill value
    P = np.array(list(position)).astype(int)  # position coordinates(numpy array)
    Rs = np.array(list(R.shape)).astype(int)  # output block dimensions (numpy array)
    Zs = np.array(list(Z.shape)).astype(int)  # original volume dimensions (numpy array)

    R_start = np.zeros(len(shape)).astype(int)
    R_stop = np.array(list(shape)).astype(int)
    Z_start = P - Rs // 2
    Z_start_cor = (np.maximum(Z_start, 0)).tolist()  # handle borders
    R_start = R_start + (Z_start_cor - Z_start)
    Z_stop = (P + Rs // 2) + Rs % 2
    Z_stop_cor = (np.minimum(Z_stop, Zs)).tolist()  # handle borders
    R_stop = R_stop - (Z_stop - Z_stop_cor)

    R[R_start[0] : R_stop[0], R_start[1] : R_stop[1], R_start[2] : R_stop[2]] = Z[
        Z_start_cor[0] : Z_stop_cor[0],
        Z_start_cor[1] : Z_stop_cor[1],
        Z_start_cor[2] : Z_stop_cor[2],
    ]

    return R


class ParcellateInputSpec(BaseInterfaceInputSpec):
    subject_id = traits.String(mandatory=True, desc="Subject ID")
    parcellation_name = traits.Enum(
        "scale500",
        ["scale33", "scale60", "scale125", "scale250", "scale500"],
        usedefault=True,
    )
    freesurfer_dir = Directory(exists=True, desc="Freesurfer main directory")
    subjects_dir = Directory(exists=True, desc="Freesurfer subjects directory")
    out_roi_file = File(
        genfile=True, desc="Region of Interest file for connectivity mapping"
    )
    dilation = traits.Bool(
        False,
        usedefault=True,
        desc="Dilate cortical parcels? Useful for fMRI connectivity",
    )


class ParcellateOutputSpec(TraitedSpec):
    roi_file = File(
        exists=True, desc="Region of Interest file for connectivity mapping"
    )
    roiv_file = File(desc="Region of Interest file for fMRI connectivity mapping")
    white_matter_mask_file = File(exists=True, desc="White matter mask file")
    cc_unknown_file = File(
        desc="Image file with regions labelled as unknown cortical structures",
        exists=True,
    )
    ribbon_file = File(desc="Image file detailing the cortical ribbon", exists=True)
    aseg_file = File(
        desc='Automated segmentation file converted from Freesurfer "subjects" directory',
        exists=True,
    )
    roi_file_in_structural_space = File(
        desc="ROI image resliced to the dimensions of the original structural image",
        exists=True,
    )
    dilated_roi_file_in_structural_space = File(
        desc="dilated ROI image resliced to the dimensions of the original structural image"
    )


class Parcellate(LibraryBaseInterface):
    """Subdivides segmented ROI file into smaller subregions

    This interface implements the same procedure as in the ConnectomeMapper's
    parcellation stage (cmp/stages/parcellation/maskcreation.py) for a single
    parcellation scheme (e.g. 'scale500').

    Example
    -------

    >>> import nipype.interfaces.cmtk as cmtk
    >>> parcellate = cmtk.Parcellate()
    >>> parcellate.inputs.freesurfer_dir = '.'
    >>> parcellate.inputs.subjects_dir = '.'
    >>> parcellate.inputs.subject_id = 'subj1'
    >>> parcellate.inputs.dilation = True
    >>> parcellate.inputs.parcellation_name = 'scale500'
    >>> parcellate.run()                 # doctest: +SKIP
    """

    input_spec = ParcellateInputSpec
    output_spec = ParcellateOutputSpec
    _pkg = "cmp"
    imports = ("scipy",)

    def _run_interface(self, runtime):
        if self.inputs.subjects_dir:
            os.environ.update({"SUBJECTS_DIR": self.inputs.subjects_dir})

        if not os.path.exists(
            op.join(self.inputs.subjects_dir, self.inputs.subject_id)
        ):
            raise Exception
        iflogger.info("ROI_HR_th.nii.gz / fsmask_1mm.nii.gz CREATION")
        iflogger.info("=============================================")
        create_annot_label(
            self.inputs.subject_id,
            self.inputs.subjects_dir,
            self.inputs.freesurfer_dir,
            self.inputs.parcellation_name,
        )
        create_roi(
            self.inputs.subject_id,
            self.inputs.subjects_dir,
            self.inputs.freesurfer_dir,
            self.inputs.parcellation_name,
            self.inputs.dilation,
        )
        create_wm_mask(
            self.inputs.subject_id,
            self.inputs.subjects_dir,
            self.inputs.freesurfer_dir,
            self.inputs.parcellation_name,
        )
        crop_and_move_datasets(
            self.inputs.subject_id,
            self.inputs.subjects_dir,
            self.inputs.freesurfer_dir,
            self.inputs.parcellation_name,
            self.inputs.out_roi_file,
            self.inputs.dilation,
        )
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        if isdefined(self.inputs.out_roi_file):
            outputs["roi_file"] = op.abspath(self.inputs.out_roi_file)
        else:
            outputs["roi_file"] = op.abspath(self._gen_outfilename("nii.gz", "ROI"))
        if self.inputs.dilation is True:
            outputs["roiv_file"] = op.abspath(self._gen_outfilename("nii.gz", "ROIv"))
        outputs["white_matter_mask_file"] = op.abspath("fsmask_1mm.nii.gz")
        outputs["cc_unknown_file"] = op.abspath("cc_unknown.nii.gz")
        outputs["ribbon_file"] = op.abspath("ribbon.nii.gz")
        outputs["aseg_file"] = op.abspath("aseg.nii.gz")
        outputs["roi_file_in_structural_space"] = op.abspath("ROI_HR_th.nii.gz")
        if self.inputs.dilation is True:
            outputs["dilated_roi_file_in_structural_space"] = op.abspath(
                "ROIv_HR_th.nii.gz"
            )
        return outputs

    def _gen_outfilename(self, ext, prefix="ROI"):
        return prefix + "_" + self.inputs.parcellation_name + "." + ext



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dwi import DwiTool


def test_DwiTool_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        b0_file=dict(
            argstr="-b0 %s",
            extensions=None,
            position=4,
        ),
        ball_flag=dict(
            argstr="-ball",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag",
                "dti_flag2",
                "ballv_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        ballv_flag=dict(
            argstr="-ballv",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag",
                "dti_flag2",
                "ball_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        bval_file=dict(
            argstr="-bval %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        bvec_file=dict(
            argstr="-bvec %s",
            extensions=None,
            position=3,
        ),
        diso_val=dict(
            argstr="-diso %f",
        ),
        dpr_val=dict(
            argstr="-dpr %f",
        ),
        dti_flag=dict(
            argstr="-dti",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag2",
                "ball_flag",
                "ballv_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        dti_flag2=dict(
            argstr="-dti2",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag",
                "ball_flag",
                "ballv_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        famap_file=dict(
            argstr="-famap %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_famap.nii.gz",
        ),
        ivim_flag=dict(
            argstr="-ivim",
            position=6,
            xor=[
                "mono_flag",
                "dti_flag",
                "dti_flag2",
                "ball_flag",
                "ballv_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        logdti_file=dict(
            argstr="-logdti2 %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_logdti2.nii.gz",
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
            position=5,
        ),
        mcmap_file=dict(
            argstr="-mcmap %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_mcmap.nii.gz",
        ),
        mdmap_file=dict(
            argstr="-mdmap %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_mdmap.nii.gz",
        ),
        mono_flag=dict(
            argstr="-mono",
            position=6,
            xor=[
                "ivim_flag",
                "dti_flag",
                "dti_flag2",
                "ball_flag",
                "ballv_flag",
                "nod_flag",
                "nodv_flag",
            ],
        ),
        nod_flag=dict(
            argstr="-nod",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag",
                "dti_flag2",
                "ball_flag",
                "ballv_flag",
                "nodv_flag",
            ],
        ),
        nodv_flag=dict(
            argstr="-nodv",
            position=6,
            xor=[
                "mono_flag",
                "ivim_flag",
                "dti_flag",
                "dti_flag2",
                "ball_flag",
                "ballv_flag",
                "nod_flag",
            ],
        ),
        rgbmap_file=dict(
            argstr="-rgbmap %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_rgbmap.nii.gz",
        ),
        source_file=dict(
            argstr="-source %s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        syn_file=dict(
            argstr="-syn %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_syn.nii.gz",
            requires=["bvec_file", "b0_file"],
        ),
        v1map_file=dict(
            argstr="-v1map %s",
            extensions=None,
            name_source=["source_file"],
            name_template="%s_v1map.nii.gz",
        ),
    )
    inputs = DwiTool.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DwiTool_outputs():
    output_map = dict(
        famap_file=dict(
            extensions=None,
        ),
        logdti_file=dict(
            extensions=None,
        ),
        mcmap_file=dict(
            extensions=None,
        ),
        mdmap_file=dict(
            extensions=None,
        ),
        rgbmap_file=dict(
            extensions=None,
        ),
        syn_file=dict(
            extensions=None,
        ),
        v1map_file=dict(
            extensions=None,
        ),
    )
    outputs = DwiTool.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""DTITK registration interfaces

DTI-TK developed by Gary Hui Zhang, gary.zhang@ucl.ac.uk
For additional help, visit http://dti-tk.sf.net

The high-dimensional tensor-based DTI registration algorithm

Zhang, H., Avants, B.B, Yushkevich, P.A., Woo, J.H., Wang, S., McCluskey, L.H.,
 Elman, L.B., Melhem, E.R., Gee, J.C., High-dimensional spatial normalization
 of diffusion tensor images improves the detection of white matter differences
 in amyotrophic lateral sclerosis, IEEE Transactions on Medical Imaging,
 26(11):1585-1597, November 2007. PMID: 18041273.

The original piecewise-affine tensor-based DTI registration algorithm at the
core of DTI-TK

Zhang, H., Yushkevich, P.A., Alexander, D.C., Gee, J.C., Deformable
 registration of diffusion tensor MR images with explicit orientation
 optimization, Medical Image Analysis, 10(5):764-785, October 2006. PMID:
 16899392.

"""

from ..base import TraitedSpec, CommandLineInputSpec, traits, File, isdefined
from ...utils.filemanip import fname_presuffix, split_filename
from .base import CommandLineDtitk, DTITKRenameMixin
import os

__docformat__ = "restructuredtext"


class RigidInputSpec(CommandLineInputSpec):
    fixed_file = File(
        desc="fixed tensor volume",
        exists=True,
        mandatory=True,
        position=0,
        argstr="%s",
        copyfile=False,
    )
    moving_file = File(
        desc="moving tensor volume",
        exists=True,
        mandatory=True,
        position=1,
        argstr="%s",
        copyfile=False,
    )
    similarity_metric = traits.Enum(
        "EDS",
        "GDS",
        "DDS",
        "NMI",
        mandatory=True,
        position=2,
        argstr="%s",
        desc="similarity metric",
        usedefault=True,
    )
    sampling_xyz = traits.Tuple(
        (4, 4, 4),
        mandatory=True,
        position=3,
        argstr="%g %g %g",
        usedefault=True,
        desc="dist between samp points (mm) (x,y,z)",
    )
    ftol = traits.Float(
        mandatory=True,
        position=4,
        argstr="%g",
        desc="cost function tolerance",
        default_value=0.01,
        usedefault=True,
    )
    initialize_xfm = File(
        copyfile=True,
        desc="Initialize w/DTITK-FORMAT" "affine",
        position=5,
        argstr="%s",
        exists=True,
    )


class RigidOutputSpec(TraitedSpec):
    out_file = File(exists=True)
    out_file_xfm = File(exists=True)


class Rigid(CommandLineDtitk):
    """Performs rigid registration between two tensor volumes

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.Rigid()
    >>> node.inputs.fixed_file = 'im1.nii'
    >>> node.inputs.moving_file = 'im2.nii'
    >>> node.inputs.similarity_metric = 'EDS'
    >>> node.inputs.sampling_xyz = (4,4,4)
    >>> node.inputs.ftol = 0.01
    >>> node.cmdline
    'dti_rigid_reg im1.nii im2.nii EDS 4 4 4 0.01'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = RigidInputSpec
    output_spec = RigidOutputSpec
    _cmd = "dti_rigid_reg"

    """def _format_arg(self, name, spec, value):
        if name == 'initialize_xfm':
            value = 1
        return super(Rigid, self)._format_arg(name, spec, value)"""

    def _run_interface(self, runtime):
        runtime = super(Rigid, self)._run_interface(runtime)
        if """.aff doesn't exist or can't be opened""" in runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        moving = self.inputs.moving_file
        outputs["out_file_xfm"] = fname_presuffix(moving, suffix=".aff", use_ext=False)
        outputs["out_file"] = fname_presuffix(moving, suffix="_aff")
        return outputs


class Affine(Rigid):
    """Performs affine registration between two tensor volumes

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.Affine()
    >>> node.inputs.fixed_file = 'im1.nii'
    >>> node.inputs.moving_file = 'im2.nii'
    >>> node.inputs.similarity_metric = 'EDS'
    >>> node.inputs.sampling_xyz = (4,4,4)
    >>> node.inputs.ftol = 0.01
    >>> node.inputs.initialize_xfm = 'im_affine.aff'
    >>> node.cmdline
    'dti_affine_reg im1.nii im2.nii EDS 4 4 4 0.01 im_affine.aff'
    >>> node.run() # doctest: +SKIP
    """

    _cmd = "dti_affine_reg"


class DiffeoInputSpec(CommandLineInputSpec):
    fixed_file = File(desc="fixed tensor volume", exists=True, position=0, argstr="%s")
    moving_file = File(
        desc="moving tensor volume",
        exists=True,
        position=1,
        argstr="%s",
        copyfile=False,
    )
    mask_file = File(desc="mask", exists=True, position=2, argstr="%s")
    legacy = traits.Enum(
        1,
        desc="legacy parameter; always set to 1",
        usedefault=True,
        mandatory=True,
        position=3,
        argstr="%d",
    )
    n_iters = traits.Int(
        6,
        desc="number of iterations",
        mandatory=True,
        position=4,
        argstr="%d",
        usedefault=True,
    )
    ftol = traits.Float(
        0.002,
        desc="iteration for the optimization to stop",
        mandatory=True,
        position=5,
        argstr="%g",
        usedefault=True,
    )


class DiffeoOutputSpec(TraitedSpec):
    out_file = File(exists=True)
    out_file_xfm = File(exists=True)


class Diffeo(CommandLineDtitk):
    """Performs diffeomorphic registration between two tensor volumes

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.Diffeo()
    >>> node.inputs.fixed_file = 'im1.nii'
    >>> node.inputs.moving_file = 'im2.nii'
    >>> node.inputs.mask_file = 'mask.nii'
    >>> node.inputs.legacy = 1
    >>> node.inputs.n_iters = 6
    >>> node.inputs.ftol = 0.002
    >>> node.cmdline
    'dti_diffeomorphic_reg im1.nii im2.nii mask.nii 1 6 0.002'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = DiffeoInputSpec
    output_spec = DiffeoOutputSpec
    _cmd = "dti_diffeomorphic_reg"

    def _list_outputs(self):
        outputs = self.output_spec().get()
        moving = self.inputs.moving_file
        outputs["out_file_xfm"] = fname_presuffix(moving, suffix="_diffeo.df")
        outputs["out_file"] = fname_presuffix(moving, suffix="_diffeo")
        return outputs


class ComposeXfmInputSpec(CommandLineInputSpec):
    in_df = File(
        desc="diffeomorphic warp file", exists=True, argstr="-df %s", mandatory=True
    )
    in_aff = File(
        desc="affine transform file", exists=True, argstr="-aff %s", mandatory=True
    )
    out_file = File(desc="output path", argstr="-out %s", genfile=True)


class ComposeXfmOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class ComposeXfm(CommandLineDtitk):
    """
     Combines diffeomorphic and affine transforms

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.ComposeXfm()
    >>> node.inputs.in_df = 'im_warp.df.nii'
    >>> node.inputs.in_aff= 'im_affine.aff'
    >>> node.cmdline
    'dfRightComposeAffine -aff im_affine.aff -df im_warp.df.nii -out
     im_warp_affdf.df.nii'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = ComposeXfmInputSpec
    output_spec = ComposeXfmOutputSpec
    _cmd = "dfRightComposeAffine"

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = self._gen_filename("out_file")
        outputs["out_file"] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name != "out_file":
            return
        path, base, ext = split_filename(self.inputs.in_df)
        suffix = "_affdf"
        if base.endswith(".df"):
            suffix += ".df"
            base = base[:-3]
        return fname_presuffix(base, suffix=suffix + ext, use_ext=False)


class AffSymTensor3DVolInputSpec(CommandLineInputSpec):
    in_file = File(
        desc="moving tensor volume", exists=True, argstr="-in %s", mandatory=True
    )
    out_file = File(
        desc="output filename",
        argstr="-out %s",
        name_source="in_file",
        name_template="%s_affxfmd",
        keep_extension=True,
    )
    transform = File(
        exists=True,
        argstr="-trans %s",
        xor=["target", "translation", "euler", "deformation"],
        desc="transform to apply: specify an input transformation"
        " file; parameters input will be ignored",
    )
    interpolation = traits.Enum(
        "LEI",
        "EI",
        usedefault=True,
        argstr="-interp %s",
        desc="Log Euclidean/Euclidean Interpolation",
    )
    reorient = traits.Enum(
        "PPD",
        "NO",
        "FS",
        argstr="-reorient %s",
        usedefault=True,
        desc="Reorientation strategy: "
        "preservation of principal direction, no "
        "reorientation, or finite strain",
    )
    target = File(
        exists=True,
        argstr="-target %s",
        xor=["transform"],
        desc="output volume specification read from the target " "volume if specified",
    )
    translation = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="translation (x,y,z) in mm",
        argstr="-translation %g %g %g",
        xor=["transform"],
    )
    euler = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="(theta, phi, psi) in degrees",
        xor=["transform"],
        argstr="-euler %g %g %g",
    )
    deformation = traits.Tuple(
        (traits.Float(),) * 6,
        desc="(xx,yy,zz,xy,yz,xz)",
        xor=["transform"],
        argstr="-deformation %g %g %g %g %g %g",
    )


class AffSymTensor3DVolOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class AffSymTensor3DVol(CommandLineDtitk):
    """
    Applies affine transform to a tensor volume

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.AffSymTensor3DVol()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.transform = 'im_affine.aff'
    >>> node.cmdline
    'affineSymTensor3DVolume -in im1.nii -interp LEI -out im1_affxfmd.nii
     -reorient PPD -trans im_affine.aff'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = AffSymTensor3DVolInputSpec
    output_spec = AffSymTensor3DVolOutputSpec
    _cmd = "affineSymTensor3DVolume"


class AffScalarVolInputSpec(CommandLineInputSpec):
    in_file = File(
        desc="moving scalar volume", exists=True, argstr="-in %s", mandatory=True
    )
    out_file = File(
        desc="output filename",
        argstr="-out %s",
        name_source="in_file",
        name_template="%s_affxfmd",
        keep_extension=True,
    )
    transform = File(
        exists=True,
        argstr="-trans %s",
        xor=["target", "translation", "euler", "deformation"],
        desc="transform to apply: specify an input transformation"
        " file; parameters input will be ignored",
    )
    interpolation = traits.Enum(
        "trilinear",
        "NN",
        usedefault=True,
        argstr="-interp %s",
        desc="trilinear or nearest neighbor" " interpolation",
    )
    target = File(
        exists=True,
        argstr="-target %s",
        xor=["transform"],
        desc="output volume specification read from the target " "volume if specified",
    )
    translation = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="translation (x,y,z) in mm",
        argstr="-translation %g %g %g",
        xor=["transform"],
    )
    euler = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="(theta, phi, psi) in degrees",
        xor=["transform"],
        argstr="-euler %g %g %g",
    )
    deformation = traits.Tuple(
        (traits.Float(),) * 6,
        desc="(xx,yy,zz,xy,yz,xz)",
        xor=["transform"],
        argstr="-deformation %g %g %g %g %g %g",
    )


class AffScalarVolOutputSpec(TraitedSpec):
    out_file = File(desc="moved volume", exists=True)


class AffScalarVol(CommandLineDtitk):
    """
    Applies affine transform to a scalar volume

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.AffScalarVol()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.transform = 'im_affine.aff'
    >>> node.cmdline
    'affineScalarVolume -in im1.nii -interp 0 -out im1_affxfmd.nii -trans
     im_affine.aff'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = AffScalarVolInputSpec
    output_spec = AffScalarVolOutputSpec
    _cmd = "affineScalarVolume"

    def _format_arg(self, name, spec, value):
        if name == "interpolation":
            value = {"trilinear": 0, "NN": 1}[value]
        return super(AffScalarVol, self)._format_arg(name, spec, value)


class DiffeoSymTensor3DVolInputSpec(CommandLineInputSpec):
    in_file = File(
        desc="moving tensor volume", exists=True, argstr="-in %s", mandatory=True
    )
    out_file = File(
        desc="output filename",
        argstr="-out %s",
        name_source="in_file",
        name_template="%s_diffeoxfmd",
        keep_extension=True,
    )
    transform = File(
        exists=True, argstr="-trans %s", mandatory=True, desc="transform to apply"
    )
    df = traits.Str("FD", argstr="-df %s", usedefault=True)
    interpolation = traits.Enum(
        "LEI",
        "EI",
        usedefault=True,
        argstr="-interp %s",
        desc="Log Euclidean/Euclidean Interpolation",
    )
    reorient = traits.Enum(
        "PPD",
        "FS",
        argstr="-reorient %s",
        usedefault=True,
        desc="Reorientation strategy: "
        "preservation of principal direction or finite "
        "strain",
    )
    target = File(
        exists=True,
        argstr="-target %s",
        xor=["voxel_size"],
        desc="output volume specification read from the target " "volume if specified",
    )
    voxel_size = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="xyz voxel size (superseded by target)",
        argstr="-vsize %g %g %g",
        xor=["target"],
    )
    flip = traits.Tuple(
        (traits.Int(), traits.Int(), traits.Int()), argstr="-flip %d %d %d"
    )
    resampling_type = traits.Enum(
        "backward",
        "forward",
        desc="use backward or forward resampling",
        argstr="-type %s",
    )


class DiffeoSymTensor3DVolOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class DiffeoSymTensor3DVol(CommandLineDtitk):
    """
    Applies diffeomorphic transform to a tensor volume

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.DiffeoSymTensor3DVol()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.transform = 'im_warp.df.nii'
    >>> node.cmdline
    'deformationSymTensor3DVolume -df FD -in im1.nii -interp LEI -out
     im1_diffeoxfmd.nii -reorient PPD -trans im_warp.df.nii'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = DiffeoSymTensor3DVolInputSpec
    output_spec = DiffeoSymTensor3DVolOutputSpec
    _cmd = "deformationSymTensor3DVolume"

    def _format_arg(self, name, spec, value):
        if name == "resampling_type":
            value = {"forward": 0, "backward": 1}[value]
        return super(DiffeoSymTensor3DVol, self)._format_arg(name, spec, value)


class DiffeoScalarVolInputSpec(CommandLineInputSpec):
    in_file = File(
        desc="moving scalar volume", exists=True, argstr="-in %s", mandatory=True
    )
    out_file = File(
        desc="output filename",
        argstr="-out %s",
        name_source="in_file",
        name_template="%s_diffeoxfmd",
        keep_extension=True,
    )
    transform = File(
        exists=True, argstr="-trans %s", mandatory=True, desc="transform to apply"
    )
    target = File(
        exists=True,
        argstr="-target %s",
        xor=["voxel_size"],
        desc="output volume specification read from the target " "volume if specified",
    )
    voxel_size = traits.Tuple(
        (traits.Float(), traits.Float(), traits.Float()),
        desc="xyz voxel size (superseded by target)",
        argstr="-vsize %g %g %g",
        xor=["target"],
    )
    flip = traits.Tuple(
        (traits.Int(), traits.Int(), traits.Int()), argstr="-flip %d %d %d"
    )
    resampling_type = traits.Enum(
        "backward",
        "forward",
        desc="use backward or forward resampling",
        argstr="-type %s",
    )
    interpolation = traits.Enum(
        "trilinear",
        "NN",
        desc="trilinear, or nearest neighbor",
        argstr="-interp %s",
        usedefault=True,
    )


class DiffeoScalarVolOutputSpec(TraitedSpec):
    out_file = File(desc="moved volume", exists=True)


class DiffeoScalarVol(CommandLineDtitk):
    """
    Applies diffeomorphic transform to a scalar volume

    Example
    -------

    >>> from nipype.interfaces import dtitk
    >>> node = dtitk.DiffeoScalarVol()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.transform = 'im_warp.df.nii'
    >>> node.cmdline
    'deformationScalarVolume -in im1.nii -interp 0 -out im1_diffeoxfmd.nii
     -trans im_warp.df.nii'
    >>> node.run() # doctest: +SKIP
    """

    input_spec = DiffeoScalarVolInputSpec
    output_spec = DiffeoScalarVolOutputSpec
    _cmd = "deformationScalarVolume"

    def _format_arg(self, name, spec, value):
        if name == "resampling_type":
            value = {"forward": 0, "backward": 1}[value]
        elif name == "interpolation":
            value = {"trilinear": 0, "NN": 1}[value]
        return super(DiffeoScalarVol, self)._format_arg(name, spec, value)


class RigidTask(DTITKRenameMixin, Rigid):
    pass


class AffineTask(DTITKRenameMixin, Affine):
    pass


class DiffeoTask(DTITKRenameMixin, Diffeo):
    pass


class ComposeXfmTask(DTITKRenameMixin, ComposeXfm):
    pass


class affScalarVolTask(DTITKRenameMixin, AffScalarVol):
    pass


class affSymTensor3DVolTask(DTITKRenameMixin, AffSymTensor3DVol):
    pass


class diffeoScalarVolTask(DTITKRenameMixin, DiffeoScalarVol):
    pass


class diffeoSymTensor3DVolTask(DTITKRenameMixin, DiffeoSymTensor3DVol):
    pass



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..cmtk import ROIGen


def test_ROIGen_inputs():
    input_map = dict(
        LUT_file=dict(
            extensions=None,
            xor=["use_freesurfer_LUT"],
        ),
        aparc_aseg_file=dict(
            extensions=None,
            mandatory=True,
        ),
        freesurfer_dir=dict(
            requires=["use_freesurfer_LUT"],
        ),
        out_dict_file=dict(
            extensions=None,
            genfile=True,
        ),
        out_roi_file=dict(
            extensions=None,
            genfile=True,
        ),
        use_freesurfer_LUT=dict(
            xor=["LUT_file"],
        ),
    )
    inputs = ROIGen.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ROIGen_outputs():
    output_map = dict(
        dict_file=dict(
            extensions=None,
        ),
        roi_file=dict(
            extensions=None,
        ),
    )
    outputs = ROIGen.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import os
from ...base import Undefined
from ..model import Level1Design


def test_level1design(tmpdir):
    old = tmpdir.chdir()
    l = Level1Design()
    runinfo = dict(
        cond=[{"name": "test_condition", "onset": [0, 10], "duration": [10, 10]}],
        regress=[],
    )
    runidx = 0
    contrasts = Undefined
    do_tempfilter = False
    orthogonalization = {}
    basic_ev_parameters = {"temporalderiv": False}
    convolution_variants = [
        ("custom", 7, {"temporalderiv": False, "bfcustompath": "/some/path"}),
        ("hrf", 3, basic_ev_parameters),
        ("dgamma", 3, basic_ev_parameters),
        ("gamma", 2, basic_ev_parameters),
        ("none", 0, basic_ev_parameters),
    ]
    for key, val, ev_parameters in convolution_variants:
        output_num, output_txt = Level1Design._create_ev_files(
            l,
            os.getcwd(),
            runinfo,
            runidx,
            ev_parameters,
            orthogonalization,
            contrasts,
            do_tempfilter,
            key,
        )
        assert "set fmri(convolve1) {0}".format(val) in output_txt



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import FitGLM


def test_FitGLM_inputs():
    input_map = dict(
        TR=dict(
            mandatory=True,
        ),
        drift_model=dict(
            usedefault=True,
        ),
        hrf_model=dict(
            usedefault=True,
        ),
        mask=dict(
            extensions=None,
        ),
        method=dict(
            usedefault=True,
        ),
        model=dict(
            usedefault=True,
        ),
        normalize_design_matrix=dict(
            usedefault=True,
        ),
        plot_design_matrix=dict(
            usedefault=True,
        ),
        save_residuals=dict(
            usedefault=True,
        ),
        session_info=dict(
            mandatory=True,
        ),
    )
    inputs = FitGLM.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FitGLM_outputs():
    output_map = dict(
        a=dict(
            extensions=None,
        ),
        axis=dict(),
        beta=dict(
            extensions=None,
        ),
        constants=dict(),
        dof=dict(),
        nvbeta=dict(),
        reg_names=dict(),
        residuals=dict(
            extensions=None,
        ),
        s2=dict(
            extensions=None,
        ),
    )
    outputs = FitGLM.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""ANTS Apply Transforms interface
"""
import os

from .base import ANTSCommand, ANTSCommandInputSpec
from ..base import TraitedSpec, File, traits, isdefined, InputMultiObject
from ...utils.filemanip import split_filename


class WarpTimeSeriesImageMultiTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        4, 3, argstr="%d", usedefault=True, desc="image dimension (3 or 4)", position=1
    )
    input_image = File(
        argstr="%s",
        mandatory=True,
        copyfile=True,
        desc=(
            "image to apply transformation to (generally a " "coregistered functional)"
        ),
    )
    out_postfix = traits.Str(
        "_wtsimt",
        argstr="%s",
        usedefault=True,
        desc=("Postfix that is prepended to all output " "files (default = _wtsimt)"),
    )
    reference_image = File(
        argstr="-R %s",
        xor=["tightest_box"],
        desc="reference image space that you wish to warp INTO",
    )
    tightest_box = traits.Bool(
        argstr="--tightest-bounding-box",
        desc=(
            "computes tightest bounding box (overrided by " "reference_image if given)"
        ),
        xor=["reference_image"],
    )
    reslice_by_header = traits.Bool(
        argstr="--reslice-by-header",
        desc=(
            "Uses orientation matrix and origin encoded in "
            "reference image file header. Not typically used "
            "with additional transforms"
        ),
    )
    use_nearest = traits.Bool(
        argstr="--use-NN", desc="Use nearest neighbor interpolation"
    )
    use_bspline = traits.Bool(
        argstr="--use-Bspline", desc="Use 3rd order B-Spline interpolation"
    )
    transformation_series = InputMultiObject(
        File(exists=True),
        argstr="%s",
        desc="transformation file(s) to be applied",
        mandatory=True,
        copyfile=False,
    )
    invert_affine = traits.List(
        traits.Int,
        desc=(
            "List of Affine transformations to invert."
            "E.g.: [1,4,5] inverts the 1st, 4th, and 5th Affines "
            "found in transformation_series. Note that indexing "
            "starts with 1 and does not include warp fields. Affine "
            "transformations are distinguished "
            'from warp fields by the word "affine" included in their filenames.'
        ),
    )


class WarpTimeSeriesImageMultiTransformOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="Warped image")


class WarpTimeSeriesImageMultiTransform(ANTSCommand):
    """Warps a time-series from one space to another

    Examples
    --------

    >>> from nipype.interfaces.ants import WarpTimeSeriesImageMultiTransform
    >>> wtsimt = WarpTimeSeriesImageMultiTransform()
    >>> wtsimt.inputs.input_image = 'resting.nii'
    >>> wtsimt.inputs.reference_image = 'ants_deformed.nii.gz'
    >>> wtsimt.inputs.transformation_series = ['ants_Warp.nii.gz','ants_Affine.txt']
    >>> wtsimt.cmdline
    'WarpTimeSeriesImageMultiTransform 4 resting.nii resting_wtsimt.nii -R ants_deformed.nii.gz ants_Warp.nii.gz \
ants_Affine.txt'

    >>> wtsimt = WarpTimeSeriesImageMultiTransform()
    >>> wtsimt.inputs.input_image = 'resting.nii'
    >>> wtsimt.inputs.reference_image = 'ants_deformed.nii.gz'
    >>> wtsimt.inputs.transformation_series = ['ants_Warp.nii.gz','ants_Affine.txt']
    >>> wtsimt.inputs.invert_affine = [1] # # this will invert the 1st Affine file: ants_Affine.txt
    >>> wtsimt.cmdline
    'WarpTimeSeriesImageMultiTransform 4 resting.nii resting_wtsimt.nii -R ants_deformed.nii.gz ants_Warp.nii.gz \
-i ants_Affine.txt'
    """

    _cmd = "WarpTimeSeriesImageMultiTransform"
    input_spec = WarpTimeSeriesImageMultiTransformInputSpec
    output_spec = WarpTimeSeriesImageMultiTransformOutputSpec

    def _format_arg(self, opt, spec, val):
        if opt == "out_postfix":
            _, name, ext = split_filename(os.path.abspath(self.inputs.input_image))
            return name + val + ext
        if opt == "transformation_series":
            series = []
            affine_counter = 0
            affine_invert = []
            for transformation in val:
                if "Affine" in transformation and isdefined(self.inputs.invert_affine):
                    affine_counter += 1
                    if affine_counter in self.inputs.invert_affine:
                        series += ["-i"]
                        affine_invert.append(affine_counter)
                series += [transformation]

            if isdefined(self.inputs.invert_affine):
                diff_inv = set(self.inputs.invert_affine) - set(affine_invert)
                if diff_inv:
                    raise Exceptions(
                        "Review invert_affine, not all indexes from invert_affine were used, "
                        "check the description for the full definition"
                    )

            return " ".join(series)
        return super(WarpTimeSeriesImageMultiTransform, self)._format_arg(
            opt, spec, val
        )

    def _list_outputs(self):
        outputs = self._outputs().get()
        _, name, ext = split_filename(os.path.abspath(self.inputs.input_image))
        outputs["output_image"] = os.path.join(
            os.getcwd(), "".join((name, self.inputs.out_postfix, ext))
        )
        return outputs

    def _run_interface(self, runtime, correct_return_codes=[0]):
        runtime = super(WarpTimeSeriesImageMultiTransform, self)._run_interface(
            runtime, correct_return_codes=[0, 1]
        )
        if "100 % complete" not in runtime.stdout:
            self.raise_exception(runtime)
        return runtime


class WarpImageMultiTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", usedefault=True, desc="image dimension (2 or 3)", position=1
    )
    input_image = File(
        argstr="%s",
        mandatory=True,
        desc=(
            "image to apply transformation to (generally a " "coregistered functional)"
        ),
        position=2,
    )
    output_image = File(
        genfile=True,
        hash_files=False,
        argstr="%s",
        desc="name of the output warped image",
        position=3,
        xor=["out_postfix"],
    )
    out_postfix = File(
        "_wimt",
        usedefault=True,
        hash_files=False,
        desc=("Postfix that is prepended to all output " "files (default = _wimt)"),
        xor=["output_image"],
    )
    reference_image = File(
        argstr="-R %s",
        xor=["tightest_box"],
        desc="reference image space that you wish to warp INTO",
    )
    tightest_box = traits.Bool(
        argstr="--tightest-bounding-box",
        desc=(
            "computes tightest bounding box (overrided by " "reference_image if given)"
        ),
        xor=["reference_image"],
    )
    reslice_by_header = traits.Bool(
        argstr="--reslice-by-header",
        desc=(
            "Uses orientation matrix and origin encoded in "
            "reference image file header. Not typically used "
            "with additional transforms"
        ),
    )
    use_nearest = traits.Bool(
        argstr="--use-NN", desc="Use nearest neighbor interpolation"
    )
    use_bspline = traits.Bool(
        argstr="--use-BSpline", desc="Use 3rd order B-Spline interpolation"
    )
    transformation_series = InputMultiObject(
        File(exists=True),
        argstr="%s",
        desc="transformation file(s) to be applied",
        mandatory=True,
        position=-1,
    )
    invert_affine = traits.List(
        traits.Int,
        desc=(
            "List of Affine transformations to invert."
            "E.g.: [1,4,5] inverts the 1st, 4th, and 5th Affines "
            "found in transformation_series. Note that indexing "
            "starts with 1 and does not include warp fields. Affine "
            "transformations are distinguished "
            'from warp fields by the word "affine" included in their filenames.'
        ),
    )


class WarpImageMultiTransformOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="Warped image")


class WarpImageMultiTransform(ANTSCommand):
    """Warps an image from one space to another

    Examples
    --------

    >>> from nipype.interfaces.ants import WarpImageMultiTransform
    >>> wimt = WarpImageMultiTransform()
    >>> wimt.inputs.input_image = 'structural.nii'
    >>> wimt.inputs.reference_image = 'ants_deformed.nii.gz'
    >>> wimt.inputs.transformation_series = ['ants_Warp.nii.gz','ants_Affine.txt']
    >>> wimt.cmdline
    'WarpImageMultiTransform 3 structural.nii structural_wimt.nii -R ants_deformed.nii.gz ants_Warp.nii.gz \
ants_Affine.txt'

    >>> wimt = WarpImageMultiTransform()
    >>> wimt.inputs.input_image = 'diffusion_weighted.nii'
    >>> wimt.inputs.reference_image = 'functional.nii'
    >>> wimt.inputs.transformation_series = ['func2anat_coreg_Affine.txt','func2anat_InverseWarp.nii.gz', \
    'dwi2anat_Warp.nii.gz','dwi2anat_coreg_Affine.txt']
    >>> wimt.inputs.invert_affine = [1]  # this will invert the 1st Affine file: 'func2anat_coreg_Affine.txt'
    >>> wimt.cmdline
    'WarpImageMultiTransform 3 diffusion_weighted.nii diffusion_weighted_wimt.nii -R functional.nii \
-i func2anat_coreg_Affine.txt func2anat_InverseWarp.nii.gz dwi2anat_Warp.nii.gz dwi2anat_coreg_Affine.txt'

    """

    _cmd = "WarpImageMultiTransform"
    input_spec = WarpImageMultiTransformInputSpec
    output_spec = WarpImageMultiTransformOutputSpec

    def _gen_filename(self, name):
        if name == "output_image":
            _, name, ext = split_filename(os.path.abspath(self.inputs.input_image))
            return "".join((name, self.inputs.out_postfix, ext))
        return None

    def _format_arg(self, opt, spec, val):
        if opt == "transformation_series":
            series = []
            affine_counter = 0
            affine_invert = []
            for transformation in val:
                if "affine" in transformation.lower() and isdefined(
                    self.inputs.invert_affine
                ):
                    affine_counter += 1
                    if affine_counter in self.inputs.invert_affine:
                        series += ["-i"]
                        affine_invert.append(affine_counter)
                series += [transformation]

            if isdefined(self.inputs.invert_affine):
                diff_inv = set(self.inputs.invert_affine) - set(affine_invert)
                if diff_inv:
                    raise Exceptions(
                        "Review invert_affine, not all indexes from invert_affine were used, "
                        "check the description for the full definition"
                    )

            return " ".join(series)

        return super(WarpImageMultiTransform, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        if isdefined(self.inputs.output_image):
            outputs["output_image"] = os.path.abspath(self.inputs.output_image)
        else:
            outputs["output_image"] = os.path.abspath(
                self._gen_filename("output_image")
            )
        return outputs


class ApplyTransformsInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        2,
        3,
        4,
        argstr="--dimensionality %d",
        desc=(
            "This option forces the image to be treated "
            "as a specified-dimensional image. If not "
            "specified, antsWarp tries to infer the "
            "dimensionality from the input image."
        ),
    )
    input_image_type = traits.Enum(
        0,
        1,
        2,
        3,
        argstr="--input-image-type %d",
        desc=(
            "Option specifying the input image "
            "type of scalar (default), vector, "
            "tensor, or time series."
        ),
    )
    input_image = File(
        argstr="--input %s",
        mandatory=True,
        desc=(
            "image to apply transformation to (generally a " "coregistered functional)"
        ),
        exists=True,
    )
    output_image = traits.Str(
        argstr="--output %s", desc="output file name", genfile=True, hash_files=False
    )
    out_postfix = traits.Str(
        "_trans",
        usedefault=True,
        desc=("Postfix that is appended to all output " "files (default = _trans)"),
    )
    reference_image = File(
        argstr="--reference-image %s",
        mandatory=True,
        desc="reference image space that you wish to warp INTO",
        exists=True,
    )
    interpolation = traits.Enum(
        "Linear",
        "NearestNeighbor",
        "CosineWindowedSinc",
        "WelchWindowedSinc",
        "HammingWindowedSinc",
        "LanczosWindowedSinc",
        "MultiLabel",
        "Gaussian",
        "BSpline",
        argstr="%s",
        usedefault=True,
    )
    interpolation_parameters = traits.Either(
        traits.Tuple(traits.Int()),  # BSpline (order)
        traits.Tuple(
            traits.Float(), traits.Float()  # Gaussian/MultiLabel (sigma, alpha)
        ),
    )
    transforms = InputMultiObject(
        traits.Either(File(exists=True), "identity"),
        argstr="%s",
        mandatory=True,
        desc="transform files: will be applied in reverse order. For "
        "example, the last specified transform will be applied first.",
    )
    invert_transform_flags = InputMultiObject(traits.Bool())
    default_value = traits.Float(0.0, argstr="--default-value %g", usedefault=True)
    print_out_composite_warp_file = traits.Bool(
        False,
        requires=["output_image"],
        desc="output a composite warp file instead of a transformed image",
    )
    float = traits.Bool(
        argstr="--float %d",
        default_value=False,
        usedefault=True,
        desc="Use float instead of double for computations.",
    )


class ApplyTransformsOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="Warped image")


class ApplyTransforms(ANTSCommand):
    """ApplyTransforms, applied to an input image, transforms it according to a
    reference image and a transform (or a set of transforms).

    Examples
    --------

    >>> from nipype.interfaces.ants import ApplyTransforms
    >>> at = ApplyTransforms()
    >>> at.inputs.input_image = 'moving1.nii'
    >>> at.inputs.reference_image = 'fixed1.nii'
    >>> at.inputs.transforms = 'identity'
    >>> at.cmdline
    'antsApplyTransforms --default-value 0 --float 0 --input moving1.nii \
--interpolation Linear --output moving1_trans.nii \
--reference-image fixed1.nii --transform identity'

    >>> at = ApplyTransforms()
    >>> at.inputs.dimension = 3
    >>> at.inputs.input_image = 'moving1.nii'
    >>> at.inputs.reference_image = 'fixed1.nii'
    >>> at.inputs.output_image = 'deformed_moving1.nii'
    >>> at.inputs.interpolation = 'Linear'
    >>> at.inputs.default_value = 0
    >>> at.inputs.transforms = ['ants_Warp.nii.gz', 'trans.mat']
    >>> at.inputs.invert_transform_flags = [False, True]
    >>> at.cmdline
    'antsApplyTransforms --default-value 0 --dimensionality 3 --float 0 --input moving1.nii \
--interpolation Linear --output deformed_moving1.nii --reference-image fixed1.nii \
--transform ants_Warp.nii.gz --transform [ trans.mat, 1 ]'

    >>> at1 = ApplyTransforms()
    >>> at1.inputs.dimension = 3
    >>> at1.inputs.input_image = 'moving1.nii'
    >>> at1.inputs.reference_image = 'fixed1.nii'
    >>> at1.inputs.output_image = 'deformed_moving1.nii'
    >>> at1.inputs.interpolation = 'BSpline'
    >>> at1.inputs.interpolation_parameters = (5,)
    >>> at1.inputs.default_value = 0
    >>> at1.inputs.transforms = ['ants_Warp.nii.gz', 'trans.mat']
    >>> at1.inputs.invert_transform_flags = [False, False]
    >>> at1.cmdline
    'antsApplyTransforms --default-value 0 --dimensionality 3 --float 0 --input moving1.nii \
--interpolation BSpline[ 5 ] --output deformed_moving1.nii --reference-image fixed1.nii \
--transform ants_Warp.nii.gz --transform trans.mat'

    Identity transforms may be used as part of a chain:

    >>> at2 = ApplyTransforms()
    >>> at2.inputs.dimension = 3
    >>> at2.inputs.input_image = 'moving1.nii'
    >>> at2.inputs.reference_image = 'fixed1.nii'
    >>> at2.inputs.output_image = 'deformed_moving1.nii'
    >>> at2.inputs.interpolation = 'BSpline'
    >>> at2.inputs.interpolation_parameters = (5,)
    >>> at2.inputs.default_value = 0
    >>> at2.inputs.transforms = ['identity', 'ants_Warp.nii.gz', 'trans.mat']
    >>> at2.cmdline
    'antsApplyTransforms --default-value 0 --dimensionality 3 --float 0 --input moving1.nii \
--interpolation BSpline[ 5 ] --output deformed_moving1.nii --reference-image fixed1.nii \
--transform identity --transform ants_Warp.nii.gz --transform trans.mat'
    """

    _cmd = "antsApplyTransforms"
    input_spec = ApplyTransformsInputSpec
    output_spec = ApplyTransformsOutputSpec

    def _gen_filename(self, name):
        if name == "output_image":
            output = self.inputs.output_image
            if not isdefined(output):
                _, name, ext = split_filename(self.inputs.input_image)
                output = name + self.inputs.out_postfix + ext
            return output
        return None

    def _get_transform_filenames(self):
        retval = []
        invert_flags = self.inputs.invert_transform_flags
        if not isdefined(invert_flags):
            invert_flags = [False] * len(self.inputs.transforms)
        elif len(self.inputs.transforms) != len(invert_flags):
            raise ValueError(
                "ERROR: The invert_transform_flags list must have the same number "
                "of entries as the transforms list."
            )

        for transform, invert in zip(self.inputs.transforms, invert_flags):
            if invert:
                retval.append(f"--transform [ {transform}, 1 ]")
            else:
                retval.append(f"--transform {transform}")
        return " ".join(retval)

    def _get_output_warped_filename(self):
        if isdefined(self.inputs.print_out_composite_warp_file):
            return "--output [ %s, %d ]" % (
                self._gen_filename("output_image"),
                int(self.inputs.print_out_composite_warp_file),
            )
        else:
            return "--output %s" % (self._gen_filename("output_image"))

    def _format_arg(self, opt, spec, val):
        if opt == "output_image":
            return self._get_output_warped_filename()
        elif opt == "transforms":
            return self._get_transform_filenames()
        elif opt == "interpolation":
            if (
                self.inputs.interpolation
                in [
                    "BSpline",
                    "MultiLabel",
                    "Gaussian",
                ]
                and isdefined(self.inputs.interpolation_parameters)
            ):
                return "--interpolation %s[ %s ]" % (
                    self.inputs.interpolation,
                    ", ".join(
                        [str(param) for param in self.inputs.interpolation_parameters]
                    ),
                )
            else:
                return "--interpolation %s" % self.inputs.interpolation
        return super(ApplyTransforms, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output_image"] = os.path.abspath(self._gen_filename("output_image"))
        return outputs


class ApplyTransformsToPointsInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        2,
        3,
        4,
        argstr="--dimensionality %d",
        desc=(
            "This option forces the image to be treated "
            "as a specified-dimensional image. If not "
            "specified, antsWarp tries to infer the "
            "dimensionality from the input image."
        ),
    )
    input_file = File(
        argstr="--input %s",
        mandatory=True,
        desc=(
            "Currently, the only input supported is a csv file with"
            " columns including x,y (2D), x,y,z (3D) or x,y,z,t,label (4D) column headers."
            " The points should be defined in physical space."
            " If in doubt how to convert coordinates from your files to the space"
            " required by antsApplyTransformsToPoints try creating/drawing a simple"
            " label volume with only one voxel set to 1 and all others set to 0."
            " Write down the voxel coordinates. Then use ImageMaths LabelStats to find"
            " out what coordinates for this voxel antsApplyTransformsToPoints is"
            " expecting."
        ),
        exists=True,
    )
    output_file = traits.Str(
        argstr="--output %s",
        desc="Name of the output CSV file",
        name_source=["input_file"],
        hash_files=False,
        name_template="%s_transformed.csv",
    )
    transforms = traits.List(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        desc="transforms that will be applied to the points",
    )
    invert_transform_flags = traits.List(
        traits.Bool(), desc="list indicating if a transform should be reversed"
    )


class ApplyTransformsToPointsOutputSpec(TraitedSpec):
    output_file = File(exists=True, desc="csv file with transformed coordinates")


class ApplyTransformsToPoints(ANTSCommand):
    """ApplyTransformsToPoints, applied to an CSV file, transforms coordinates
    using provided transform (or a set of transforms).

    Examples
    --------

    >>> from nipype.interfaces.ants import ApplyTransforms
    >>> at = ApplyTransformsToPoints()
    >>> at.inputs.dimension = 3
    >>> at.inputs.input_file = 'moving.csv'
    >>> at.inputs.transforms = ['trans.mat', 'ants_Warp.nii.gz']
    >>> at.inputs.invert_transform_flags = [False, False]
    >>> at.cmdline
    'antsApplyTransformsToPoints --dimensionality 3 --input moving.csv --output moving_transformed.csv \
--transform [ trans.mat, 0 ] --transform [ ants_Warp.nii.gz, 0 ]'


    """

    _cmd = "antsApplyTransformsToPoints"
    input_spec = ApplyTransformsToPointsInputSpec
    output_spec = ApplyTransformsToPointsOutputSpec

    def _get_transform_filenames(self):
        retval = []
        for ii in range(len(self.inputs.transforms)):
            if isdefined(self.inputs.invert_transform_flags):
                if len(self.inputs.transforms) == len(
                    self.inputs.invert_transform_flags
                ):
                    invert_code = 1 if self.inputs.invert_transform_flags[ii] else 0
                    retval.append(
                        "--transform [ %s, %d ]"
                        % (self.inputs.transforms[ii], invert_code)
                    )
                else:
                    raise Exception(
                        (
                            "ERROR: The useInverse list must have the same number "
                            "of entries as the transformsFileName list."
                        )
                    )
            else:
                retval.append("--transform %s" % self.inputs.transforms[ii])
        return " ".join(retval)

    def _format_arg(self, opt, spec, val):
        if opt == "transforms":
            return self._get_transform_filenames()
        return super(ApplyTransformsToPoints, self)._format_arg(opt, spec, val)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dti import DTIFit


def test_DTIFit_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bgmask=dict(
            argstr="-bgmask %s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        non_linear=dict(
            argstr="-nonlinear",
            position=3,
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        scheme_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
    )
    inputs = DTIFit.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DTIFit_outputs():
    output_map = dict(
        tensor_fitted=dict(
            extensions=None,
        ),
    )
    outputs = DTIFit.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft = python sts = 4 ts = 4 sw = 4 et:
"""
AFNI modeling interfaces.

Examples
--------
See the docstrings of the individual classes for examples.

"""

import os

from ..base import (
    CommandLineInputSpec,
    CommandLine,
    Directory,
    TraitedSpec,
    traits,
    isdefined,
    File,
    InputMultiPath,
    Undefined,
    Str,
)
from ...external.due import BibTeX

from .base import (
    AFNICommandBase,
    AFNICommand,
    AFNICommandInputSpec,
    AFNICommandOutputSpec,
    Info,
)


class DeconvolveInputSpec(AFNICommandInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        desc="filenames of 3D+time input datasets. More than one filename can "
        "be given and the datasets will be auto-catenated in time. "
        "You can input a 1D time series file here, but the time axis "
        "should run along the ROW direction, not the COLUMN direction as "
        "in the 'input1D' option.",
        argstr="-input %s",
        copyfile=False,
        sep=" ",
        position=1,
    )
    sat = traits.Bool(
        desc="check the dataset time series for initial saturation transients,"
        " which should normally have been excised before data analysis.",
        argstr="-sat",
        xor=["trans"],
    )
    trans = traits.Bool(
        desc="check the dataset time series for initial saturation transients,"
        " which should normally have been excised before data analysis.",
        argstr="-trans",
        xor=["sat"],
    )
    noblock = traits.Bool(
        desc="normally, if you input multiple datasets with 'input', then "
        "the separate datasets are taken to be separate image runs that "
        "get separate baseline models. Use this options if you want to "
        "have the program consider these to be all one big run."
        "* If any of the input dataset has only 1 sub-brick, then this "
        "option is automatically invoked!"
        "* If the auto-catenation feature isn't used, then this option "
        "has no effect, no how, no way.",
        argstr="-noblock",
    )
    force_TR = traits.Float(
        desc="use this value instead of the TR in the 'input' "
        "dataset. (It's better to fix the input using Refit.)",
        argstr="-force_TR %f",
        position=0,
    )
    input1D = File(
        desc="filename of single (fMRI) .1D time series where time runs down "
        "the column.",
        argstr="-input1D %s",
        exists=True,
    )
    TR_1D = traits.Float(
        desc="TR to use with 'input1D'. This option has no effect if you do "
        "not also use 'input1D'.",
        argstr="-TR_1D %f",
    )
    legendre = traits.Bool(
        desc="use Legendre polynomials for null hypothesis (baseline model)",
        argstr="-legendre",
    )
    nolegendre = traits.Bool(
        desc="use power polynomials for null hypotheses. Don't do this "
        "unless you are crazy!",
        argstr="-nolegendre",
    )
    nodmbase = traits.Bool(
        desc="don't de-mean baseline time series", argstr="-nodmbase"
    )
    dmbase = traits.Bool(
        desc="de-mean baseline time series (default if 'polort' >= 0)", argstr="-dmbase"
    )
    svd = traits.Bool(
        desc="use SVD instead of Gaussian elimination (default)", argstr="-svd"
    )
    nosvd = traits.Bool(desc="use Gaussian elimination instead of SVD", argstr="-nosvd")
    rmsmin = traits.Float(
        desc="minimum rms error to reject reduced model (default = 0; don't "
        "use this option normally!)",
        argstr="-rmsmin %f",
    )
    nocond = traits.Bool(
        desc="DON'T calculate matrix condition number", argstr="-nocond"
    )
    singvals = traits.Bool(
        desc="print out the matrix singular values", argstr="-singvals"
    )
    goforit = traits.Int(
        desc="use this to proceed even if the matrix has bad problems (e.g., "
        "duplicate columns, large condition number, etc.).",
        argstr="-GOFORIT %i",
    )
    allzero_OK = traits.Bool(
        desc="don't consider all zero matrix columns to be the type of error "
        "that 'gotforit' is needed to ignore.",
        argstr="-allzero_OK",
    )
    dname = traits.Tuple(
        Str, Str, desc="set environmental variable to provided value", argstr="-D%s=%s"
    )
    mask = File(
        desc="filename of 3D mask dataset; only data time series from within "
        "the mask will be analyzed; results for voxels outside the mask "
        "will be set to zero.",
        argstr="-mask %s",
        exists=True,
    )
    automask = traits.Bool(
        desc="build a mask automatically from input data (will be slow for "
        "long time series datasets)",
        argstr="-automask",
    )
    STATmask = File(
        desc="build a mask from provided file, and use this mask for the "
        "purpose of reporting truncation-to float issues AND for "
        "computing the FDR curves. The actual results ARE not masked "
        "with this option (only with 'mask' or 'automask' options).",
        argstr="-STATmask %s",
        exists=True,
    )
    censor = File(
        desc="filename of censor .1D time series. This is a file of 1s and "
        "0s, indicating which time points are to be included (1) and "
        "which are to be excluded (0).",
        argstr="-censor %s",
        exists=True,
    )
    polort = traits.Int(
        desc="degree of polynomial corresponding to the null hypothesis "
        "[default: 1]",
        argstr="-polort %d",
    )
    ortvec = traits.Tuple(
        File(desc="filename", exists=True),
        Str(desc="label"),
        desc="this option lets you input a rectangular array of 1 or more "
        "baseline vectors from a file. This method is a fast way to "
        "include a lot of baseline regressors in one step. ",
        argstr="-ortvec %s %s",
    )
    x1D = File(desc="specify name for saved X matrix", argstr="-x1D %s")
    x1D_stop = traits.Bool(
        desc="stop running after writing .xmat.1D file", argstr="-x1D_stop"
    )
    cbucket = traits.Str(
        desc="Name for dataset in which to save the regression "
        "coefficients (no statistics). This dataset "
        "will be used in a -xrestore run [not yet implemented] "
        "instead of the bucket dataset, if possible.",
        argstr="-cbucket %s",
    )
    out_file = File(desc="output statistics file", argstr="-bucket %s")
    num_threads = traits.Int(
        desc="run the program with provided number of sub-processes",
        argstr="-jobs %d",
        nohash=True,
    )
    fout = traits.Bool(desc="output F-statistic for each stimulus", argstr="-fout")
    rout = traits.Bool(
        desc="output the R^2 statistic for each stimulus", argstr="-rout"
    )
    tout = traits.Bool(desc="output the T-statistic for each stimulus", argstr="-tout")
    vout = traits.Bool(
        desc="output the sample variance (MSE) for each stimulus", argstr="-vout"
    )
    nofdr = traits.Bool(
        desc="Don't compute the statistic-vs-FDR curves for the bucket " "dataset.",
        argstr="-noFDR",
    )
    global_times = traits.Bool(
        desc="use global timing for stimulus timing files",
        argstr="-global_times",
        xor=["local_times"],
    )
    local_times = traits.Bool(
        desc="use local timing for stimulus timing files",
        argstr="-local_times",
        xor=["global_times"],
    )
    num_stimts = traits.Int(
        desc="number of stimulus timing files", argstr="-num_stimts %d", position=-6
    )
    stim_times = traits.List(
        traits.Tuple(
            traits.Int(desc="k-th response model"),
            File(desc="stimulus timing file", exists=True),
            Str(desc="model"),
        ),
        desc="generate a response model from a set of stimulus times" " given in file.",
        argstr="-stim_times %d %s '%s'...",
        position=-5,
    )
    stim_label = traits.List(
        traits.Tuple(
            traits.Int(desc="k-th input stimulus"), Str(desc="stimulus label")
        ),
        desc="label for kth input stimulus (e.g., Label1)",
        argstr="-stim_label %d %s...",
        requires=["stim_times"],
        position=-4,
    )
    stim_times_subtract = traits.Float(
        desc="this option means to subtract specified seconds from each time "
        "encountered in any 'stim_times' option. The purpose of this "
        "option is to make it simple to adjust timing files for the "
        "removal of images from the start of each imaging run.",
        argstr="-stim_times_subtract %f",
    )
    num_glt = traits.Int(
        desc="number of general linear tests (i.e., contrasts)",
        argstr="-num_glt %d",
        position=-3,
    )
    gltsym = traits.List(
        Str(desc="symbolic general linear test"),
        desc="general linear tests (i.e., contrasts) using symbolic "
        "conventions (e.g., '+Label1 -Label2')",
        argstr="-gltsym 'SYM: %s'...",
        position=-2,
    )
    glt_label = traits.List(
        traits.Tuple(
            traits.Int(desc="k-th general linear test"), Str(desc="GLT label")
        ),
        desc="general linear test (i.e., contrast) labels",
        argstr="-glt_label %d %s...",
        requires=["gltsym"],
        position=-1,
    )


class DeconvolveOutputSpec(TraitedSpec):
    out_file = File(desc="output statistics file", exists=True)
    reml_script = File(
        desc="automatical generated script to run 3dREMLfit", exists=True
    )
    x1D = File(desc="save out X matrix", exists=True)
    cbucket = File(desc="output regression coefficients file (if generated)")


class Deconvolve(AFNICommand):
    """Performs OLS regression given a 4D neuroimage file and stimulus timings

    For complete details, see the `3dDeconvolve Documentation.
    <https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dDeconvolve.html>`_

    Examples
    ========

    >>> from nipype.interfaces import afni
    >>> deconvolve = afni.Deconvolve()
    >>> deconvolve.inputs.in_files = ['functional.nii', 'functional2.nii']
    >>> deconvolve.inputs.out_file = 'output.nii'
    >>> deconvolve.inputs.x1D = 'output.1D'
    >>> stim_times = [(1, 'timeseries.txt', 'SPMG1(4)')]
    >>> deconvolve.inputs.stim_times = stim_times
    >>> deconvolve.inputs.stim_label = [(1, 'Houses')]
    >>> deconvolve.inputs.gltsym = ['SYM: +Houses']
    >>> deconvolve.inputs.glt_label = [(1, 'Houses')]
    >>> deconvolve.cmdline
    "3dDeconvolve -input functional.nii functional2.nii -bucket output.nii -x1D output.1D -num_stimts 1 -stim_times 1 timeseries.txt 'SPMG1(4)' -stim_label 1 Houses -num_glt 1 -gltsym 'SYM: +Houses' -glt_label 1 Houses"
    >>> res = deconvolve.run()  # doctest: +SKIP
    """

    _cmd = "3dDeconvolve"
    input_spec = DeconvolveInputSpec
    output_spec = DeconvolveOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "gltsym":
            for n, val in enumerate(value):
                if val.startswith("SYM: "):
                    value[n] = val.lstrip("SYM: ")

        return super(Deconvolve, self)._format_arg(name, trait_spec, value)

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []
        if len(self.inputs.stim_times) and not isdefined(self.inputs.num_stimts):
            self.inputs.num_stimts = len(self.inputs.stim_times)
        if len(self.inputs.gltsym) and not isdefined(self.inputs.num_glt):
            self.inputs.num_glt = len(self.inputs.gltsym)
        if not isdefined(self.inputs.out_file):
            self.inputs.out_file = "Decon.nii"

        return super(Deconvolve, self)._parse_inputs(skip)

    def _list_outputs(self):
        outputs = self.output_spec().get()

        _gen_fname_opts = {}
        _gen_fname_opts["basename"] = self.inputs.out_file
        _gen_fname_opts["cwd"] = os.getcwd()

        if isdefined(self.inputs.x1D):
            if not self.inputs.x1D.endswith(".xmat.1D"):
                outputs["x1D"] = os.path.abspath(self.inputs.x1D + ".xmat.1D")
            else:
                outputs["x1D"] = os.path.abspath(self.inputs.x1D)
        else:
            outputs["x1D"] = self._gen_fname(suffix=".xmat.1D", **_gen_fname_opts)

        if isdefined(self.inputs.cbucket):
            outputs["cbucket"] = os.path.abspath(self.inputs.cbucket)

        outputs["reml_script"] = self._gen_fname(suffix=".REML_cmd", **_gen_fname_opts)
        # remove out_file from outputs if x1d_stop set to True
        if self.inputs.x1D_stop:
            del outputs["out_file"], outputs["cbucket"]
        else:
            outputs["out_file"] = os.path.abspath(self.inputs.out_file)

        return outputs


class RemlfitInputSpec(AFNICommandInputSpec):
    # mandatory files
    in_files = InputMultiPath(
        File(exists=True),
        desc="Read time series dataset",
        argstr='-input "%s"',
        mandatory=True,
        copyfile=False,
        sep=" ",
    )
    matrix = File(
        desc="the design matrix file, which should have been output from "
        "Deconvolve via the 'x1D' option",
        argstr="-matrix %s",
        mandatory=True,
    )
    # "Semi-Hidden Alternative Ways to Define the Matrix"
    polort = traits.Int(
        desc="if no 'matrix' option is given, AND no 'matim' option, "
        "create a matrix with Legendre polynomial regressors"
        "up to the specified order. The default value is 0, which"
        "produces a matrix with a single column of all ones",
        argstr="-polort %d",
        xor=["matrix"],
    )
    matim = File(
        desc="read a standard file as the matrix. You can use only Col as "
        "a name in GLTs with these nonstandard matrix input methods, "
        "since the other names come from the 'matrix' file. "
        "These mutually exclusive options are ignored if 'matrix' "
        "is used.",
        argstr="-matim %s",
        xor=["matrix"],
    )
    # Other arguments
    mask = File(
        desc="filename of 3D mask dataset; only data time series from within "
        "the mask will be analyzed; results for voxels outside the mask "
        "will be set to zero.",
        argstr="-mask %s",
        exists=True,
    )
    automask = traits.Bool(
        usedefault=True,
        argstr="-automask",
        desc="build a mask automatically from input data (will be slow for "
        "long time series datasets)",
    )
    STATmask = File(
        desc="filename of 3D mask dataset to be used for the purpose "
        "of reporting truncation-to float issues AND for computing the "
        "FDR curves. The actual results ARE not masked with this option "
        "(only with 'mask' or 'automask' options).",
        argstr="-STATmask %s",
        exists=True,
    )
    addbase = InputMultiPath(
        File(exists=True, desc="file containing columns to add to regression matrix"),
        desc="file(s) to add baseline model columns to the matrix with this "
        "option. Each column in the specified file(s) will be appended "
        "to the matrix. File(s) must have at least as many rows as the "
        "matrix does.",
        copyfile=False,
        sep=" ",
        argstr="-addbase %s",
    )
    slibase = InputMultiPath(
        File(exists=True, desc="file containing columns to add to regression matrix"),
        desc="similar to 'addbase' in concept, BUT each specified file "
        "must have an integer multiple of the number of slices "
        "in the input dataset(s); then, separate regression "
        "matrices are generated for each slice, with the "
        "first column of the file appended to the matrix for "
        "the first slice of the dataset, the second column of the file "
        "appended to the matrix for the first slice of the dataset, "
        "and so on. Intended to help model physiological noise in FMRI, "
        "or other effects you want to regress out that might "
        "change significantly in the inter-slice time intervals. This "
        "will slow the program down, and make it use a lot more memory "
        "(to hold all the matrix stuff).",
        argstr="-slibase %s",
    )
    slibase_sm = InputMultiPath(
        File(exists=True, desc="file containing columns to add to regression matrix"),
        desc="similar to 'slibase', BUT each file much be in slice major "
        "order (i.e. all slice0 columns come first, then all slice1 "
        "columns, etc).",
        argstr="-slibase_sm %s",
    )
    usetemp = traits.Bool(
        desc="write intermediate stuff to disk, to economize on RAM. "
        "Using this option might be necessary to run with "
        "'slibase' and with 'Grid' values above the default, "
        "since the program has to store a large number of "
        "matrices for such a problem: two for every slice and "
        "for every (a,b) pair in the ARMA parameter grid. Temporary "
        "files are written to the directory given in environment "
        "variable TMPDIR, or in /tmp, or in ./ (preference is in that "
        "order)",
        argstr="-usetemp",
    )
    nodmbase = traits.Bool(
        desc="by default, baseline columns added to the matrix via "
        "'addbase' or 'slibase' or 'dsort' will each have their "
        "mean removed (as is done in Deconvolve); this option turns this "
        "centering off",
        argstr="-nodmbase",
        requires=["addbase", "dsort"],
    )
    dsort = File(
        desc="4D dataset to be used as voxelwise baseline regressor",
        exists=True,
        copyfile=False,
        argstr="-dsort %s",
    )
    dsort_nods = traits.Bool(
        desc="if 'dsort' option is used, this command will output "
        "additional results files excluding the 'dsort' file",
        argstr="-dsort_nods",
        requires=["dsort"],
    )
    fout = traits.Bool(desc="output F-statistic for each stimulus", argstr="-fout")
    rout = traits.Bool(
        desc="output the R^2 statistic for each stimulus", argstr="-rout"
    )
    tout = traits.Bool(
        desc="output the T-statistic for each stimulus; if you use "
        "'out_file' and do not give any of 'fout', 'tout',"
        "or 'rout', then the program assumes 'fout' is activated.",
        argstr="-tout",
    )
    nofdr = traits.Bool(
        desc="do NOT add FDR curve data to bucket datasets; FDR curves can "
        "take a long time if 'tout' is used",
        argstr="-noFDR",
    )
    nobout = traits.Bool(
        desc="do NOT add baseline (null hypothesis) regressor betas "
        "to the 'rbeta_file' and/or 'obeta_file' output datasets.",
        argstr="-nobout",
    )
    gltsym = traits.List(
        traits.Either(
            traits.Tuple(File(exists=True), Str()), traits.Tuple(Str(), Str())
        ),
        desc="read a symbolic GLT from input file and associate it with a "
        "label. As in Deconvolve, you can also use the 'SYM:' method "
        "to provide the definition of the GLT directly as a string "
        "(e.g., with 'SYM: +Label1 -Label2'). Unlike Deconvolve, you "
        "MUST specify 'SYM: ' if providing the GLT directly as a "
        "string instead of from a file",
        argstr='-gltsym "%s" %s...',
    )
    out_file = File(
        desc="output dataset for beta + statistics from the REML estimation; "
        "also contains the results of any GLT analysis requested "
        "in the Deconvolve setup, similar to the 'bucket' output "
        "from Deconvolve. This dataset does NOT get the betas "
        "(or statistics) of those regressors marked as 'baseline' "
        "in the matrix file.",
        argstr="-Rbuck %s",
    )
    var_file = File(
        desc="output dataset for REML variance parameters", argstr="-Rvar %s"
    )
    rbeta_file = File(
        desc="output dataset for beta weights from the REML estimation, "
        "similar to the 'cbucket' output from Deconvolve. This dataset "
        "will contain all the beta weights, for baseline and stimulus "
        "regressors alike, unless the '-nobout' option is given -- "
        "in that case, this dataset will only get the betas for the "
        "stimulus regressors.",
        argstr="-Rbeta %s",
    )
    glt_file = File(
        desc="output dataset for beta + statistics from the REML estimation, "
        "but ONLY for the GLTs added on the REMLfit command line itself "
        "via 'gltsym'; GLTs from Deconvolve's command line will NOT "
        "be included.",
        argstr="-Rglt %s",
    )
    fitts_file = File(desc="ouput dataset for REML fitted model", argstr="-Rfitts %s")
    errts_file = File(
        desc="output dataset for REML residuals = data - fitted model",
        argstr="-Rerrts %s",
    )
    wherr_file = File(
        desc="dataset for REML residual, whitened using the estimated "
        "ARMA(1,1) correlation matrix of the noise",
        argstr="-Rwherr %s",
    )
    quiet = traits.Bool(desc="turn off most progress messages", argstr="-quiet")
    verb = traits.Bool(
        desc="turns on more progress messages, including memory usage "
        "progress reports at various stages",
        argstr="-verb",
    )
    goforit = traits.Bool(
        desc="With potential issues flagged in the design matrix, an attempt "
        "will nevertheless be made to fit the model",
        argstr="-GOFORIT",
    )
    ovar = File(
        desc="dataset for OLSQ st.dev. parameter (kind of boring)", argstr="-Ovar %s"
    )
    obeta = File(
        desc="dataset for beta weights from the OLSQ estimation", argstr="-Obeta %s"
    )
    obuck = File(
        desc="dataset for beta + statistics from the OLSQ estimation",
        argstr="-Obuck %s",
    )
    oglt = File(
        desc="dataset for beta + statistics from 'gltsym' options", argstr="-Oglt %s"
    )
    ofitts = File(desc="dataset for OLSQ fitted model", argstr="-Ofitts %s")
    oerrts = File(
        desc="dataset for OLSQ residuals (data - fitted model)", argstr="-Oerrts %s"
    )


class RemlfitOutputSpec(AFNICommandOutputSpec):
    out_file = File(
        desc="dataset for beta + statistics from the REML estimation (if " "generated"
    )
    var_file = File(desc="dataset for REML variance parameters (if generated)")
    rbeta_file = File(
        desc="dataset for beta weights from the REML estimation (if " "generated)"
    )
    rbeta_file = File(
        desc="output dataset for beta weights from the REML estimation (if " "generated"
    )
    glt_file = File(
        desc="output dataset for beta + statistics from the REML estimation, "
        "but ONLY for the GLTs added on the REMLfit command "
        "line itself via 'gltsym' (if generated)"
    )
    fitts_file = File(desc="ouput dataset for REML fitted model (if generated)")
    errts_file = File(
        desc="output dataset for REML residuals = data - fitted model (if " "generated"
    )
    wherr_file = File(
        desc="dataset for REML residual, whitened using the estimated "
        "ARMA(1,1) correlation matrix of the noise (if generated)"
    )
    ovar = File(desc="dataset for OLSQ st.dev. parameter (if generated)")
    obeta = File(
        desc="dataset for beta weights from the OLSQ estimation (if " "generated)"
    )
    obuck = File(
        desc="dataset for beta + statistics from the OLSQ estimation (if " "generated)"
    )
    oglt = File(
        desc="dataset for beta + statistics from 'gltsym' options (if " "generated"
    )
    ofitts = File(desc="dataset for OLSQ fitted model (if generated)")
    oerrts = File(
        desc="dataset for OLSQ residuals = data - fitted model (if " "generated"
    )


class Remlfit(AFNICommand):
    """Performs Generalized least squares time series fit with Restricted
    Maximum Likelihood (REML) estimation of the temporal auto-correlation
    structure.

    For complete details, see the `3dREMLfit Documentation.
    <https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dREMLfit.html>`_

    Examples
    ========

    >>> from nipype.interfaces import afni
    >>> remlfit = afni.Remlfit()
    >>> remlfit.inputs.in_files = ['functional.nii', 'functional2.nii']
    >>> remlfit.inputs.out_file = 'output.nii'
    >>> remlfit.inputs.matrix = 'output.1D'
    >>> remlfit.inputs.gltsym = [('SYM: +Lab1 -Lab2', 'TestSYM'), ('timeseries.txt', 'TestFile')]
    >>> remlfit.cmdline
    '3dREMLfit -gltsym "SYM: +Lab1 -Lab2" TestSYM -gltsym "timeseries.txt" TestFile -input "functional.nii functional2.nii" -matrix output.1D -Rbuck output.nii'
    >>> res = remlfit.run()  # doctest: +SKIP
    """

    _cmd = "3dREMLfit"
    input_spec = RemlfitInputSpec
    output_spec = RemlfitOutputSpec

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []
        return super(Remlfit, self)._parse_inputs(skip)

    def _list_outputs(self):
        outputs = self.output_spec().get()

        for key in outputs.keys():
            if isdefined(self.inputs.get()[key]):
                outputs[key] = os.path.abspath(self.inputs.get()[key])

        return outputs


class SynthesizeInputSpec(AFNICommandInputSpec):
    cbucket = File(
        desc="Read the dataset output from " "3dDeconvolve via the '-cbucket' option.",
        argstr="-cbucket %s",
        copyfile=False,
        mandatory=True,
    )
    matrix = File(
        desc="Read the matrix output from " "3dDeconvolve via the '-x1D' option.",
        argstr="-matrix %s",
        copyfile=False,
        mandatory=True,
    )
    select = traits.List(
        Str(desc="selected columns to synthesize"),
        argstr="-select %s",
        desc="A list of selected columns from the matrix (and the "
        "corresponding coefficient sub-bricks from the "
        "cbucket). Valid types include 'baseline', "
        " 'polort', 'allfunc', 'allstim', 'all', "
        "Can also provide 'something' where something matches "
        "a stim_label from 3dDeconvolve, and 'digits' where digits "
        "are the numbers of the select matrix columns by "
        "numbers (starting at 0), or number ranges of the form "
        "'3..7' and '3-7'.",
        mandatory=True,
    )
    out_file = File(
        name_template="syn",
        desc="output dataset prefix name (default 'syn')",
        argstr="-prefix %s",
    )
    dry_run = traits.Bool(
        desc="Don't compute the output, just " "check the inputs.", argstr="-dry"
    )
    TR = traits.Float(
        desc="TR to set in the output.  The default value of "
        "TR is read from the header of the matrix file.",
        argstr="-TR %f",
    )
    cenfill = traits.Enum(
        "zero",
        "nbhr",
        "none",
        argstr="-cenfill %s",
        desc="Determines how censored time points from the "
        "3dDeconvolve run will be filled. Valid types "
        "are 'zero', 'nbhr' and 'none'.",
    )


class Synthesize(AFNICommand):
    """Reads a '-cbucket' dataset and a '.xmat.1D' matrix from 3dDeconvolve,
       and synthesizes a fit dataset using user-selected sub-bricks and
       matrix columns.

    For complete details, see the `3dSynthesize Documentation.
    <https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dSynthesize.html>`_

    Examples
    ========

    >>> from nipype.interfaces import afni
    >>> synthesize = afni.Synthesize()
    >>> synthesize.inputs.cbucket = 'functional.nii'
    >>> synthesize.inputs.matrix = 'output.1D'
    >>> synthesize.inputs.select = ['baseline']
    >>> synthesize.cmdline
    '3dSynthesize -cbucket functional.nii -matrix output.1D -select baseline'
    >>> syn = synthesize.run()  # doctest: +SKIP
    """

    _cmd = "3dSynthesize"
    input_spec = SynthesizeInputSpec
    output_spec = AFNICommandOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        for key in outputs.keys():
            if isdefined(self.inputs.get()[key]):
                outputs[key] = os.path.abspath(self.inputs.get()[key])

        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..io import SSHDataGrabber


def test_SSHDataGrabber_inputs():
    input_map = dict(
        base_directory=dict(
            mandatory=True,
        ),
        download_files=dict(
            usedefault=True,
        ),
        drop_blank_outputs=dict(
            usedefault=True,
        ),
        hostname=dict(
            mandatory=True,
        ),
        password=dict(),
        raise_on_empty=dict(
            usedefault=True,
        ),
        sort_filelist=dict(
            mandatory=True,
        ),
        ssh_log_to_file=dict(
            usedefault=True,
        ),
        template=dict(
            mandatory=True,
        ),
        template_args=dict(),
        template_expression=dict(
            usedefault=True,
        ),
        username=dict(),
    )
    inputs = SSHDataGrabber.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SSHDataGrabber_outputs():
    output_map = dict()
    outputs = SSHDataGrabber.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import DicomImport


def test_DicomImport_inputs():
    input_map = dict(
        format=dict(
            field="convopts.format",
            usedefault=True,
        ),
        icedims=dict(
            field="convopts.icedims",
            usedefault=True,
        ),
        in_files=dict(
            field="data",
            mandatory=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        output_dir=dict(
            field="outdir",
            usedefault=True,
        ),
        output_dir_struct=dict(
            field="root",
            usedefault=True,
        ),
        paths=dict(),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
    )
    inputs = DicomImport.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DicomImport_outputs():
    output_map = dict(
        out_files=dict(),
    )
    outputs = DicomImport.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# coding: utf-8
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""elastix is a toolbox for rigid and nonrigid registration of images."""
from .registration import Registration, ApplyWarp, AnalyzeWarp, PointsWarp
from .utils import EditTransform



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..tensors import FindShPeaks


def test_FindShPeaks_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        directions_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        display_debug=dict(
            argstr="-debug",
        ),
        display_info=dict(
            argstr="-info",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        num_peaks=dict(
            argstr="-num %s",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            hash_files=False,
            keep_extension=False,
            name_source=["in_file"],
            name_template="%s_peak_dirs.mif",
            position=-1,
        ),
        peak_directions=dict(
            argstr="-direction %s",
            sep=" ",
        ),
        peak_threshold=dict(
            argstr="-threshold %s",
        ),
        peaks_image=dict(
            argstr="-peaks %s",
            extensions=None,
        ),
        quiet_display=dict(
            argstr="-quiet",
        ),
    )
    inputs = FindShPeaks.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FindShPeaks_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = FindShPeaks.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import pytest
from nipype.utils.functions import getsource, create_function_from_source


def _func1(x):
    return x ** 3


def test_func_to_str():
    def func1(x):
        return x ** 2

    # Should be ok with both functions!
    for f in _func1, func1:
        f_src = getsource(f)
        f_recreated = create_function_from_source(f_src)
        assert f(2.3) == f_recreated(2.3)


def test_func_to_str_err():
    bad_src = "obbledygobbledygook"
    with pytest.raises(RuntimeError):
        create_function_from_source(bad_src)


def _print_statement():
    try:
        exec('print("")')
        return True
    except SyntaxError:
        return False


def test_func_string():
    def is_string():
        return isinstance("string", str)

    wrapped_func = create_function_from_source(getsource(is_string))
    assert is_string() == wrapped_func()


def test_func_print():
    wrapped_func = create_function_from_source(getsource(_print_statement))
    assert wrapped_func()



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Module to draw an html gantt chart from logfile produced by
``nipype.utils.profiler.log_nodes_cb()``
"""
# Import packages
import sys
import random
import datetime
import simplejson as json

from collections import OrderedDict

# Pandas
try:
    import pandas as pd
except ImportError:
    print(
        "Pandas not found; in order for full functionality of this module "
        "install the pandas package"
    )
    pass


def create_event_dict(start_time, nodes_list):
    """
    Function to generate a dictionary of event (start/finish) nodes
    from the nodes list

    Parameters
    ----------
    start_time : datetime.datetime
        a datetime object of the pipeline start time
    nodes_list : list
        a list of the node dictionaries that were run in the pipeline

    Returns
    -------
    events : dictionary
        a dictionary where the key is the timedelta from the start of
        the pipeline execution to the value node it accompanies
    """

    # Import packages
    import copy

    events = {}
    for node in nodes_list:
        # Format node fields
        estimated_threads = node.get("num_threads", 1)
        estimated_memory_gb = node.get("estimated_memory_gb", 1.0)
        runtime_threads = node.get("runtime_threads", 0)
        runtime_memory_gb = node.get("runtime_memory_gb", 0.0)

        # Init and format event-based nodes
        node["estimated_threads"] = estimated_threads
        node["estimated_memory_gb"] = estimated_memory_gb
        node["runtime_threads"] = runtime_threads
        node["runtime_memory_gb"] = runtime_memory_gb
        start_node = node
        finish_node = copy.deepcopy(node)
        start_node["event"] = "start"
        finish_node["event"] = "finish"

        # Get dictionary key
        start_delta = (node["start"] - start_time).total_seconds()
        finish_delta = (node["finish"] - start_time).total_seconds()

        # Populate dictionary
        if events.get(start_delta) or events.get(finish_delta):
            err_msg = "Event logged twice or events started at exact same time!"
            raise KeyError(err_msg)
        events[start_delta] = start_node
        events[finish_delta] = finish_node

    # Return events dictionary
    return events


def log_to_dict(logfile):
    """
    Function to extract log node dictionaries into a list of python
    dictionaries and return the list as well as the final node

    Parameters
    ----------
    logfile : string
        path to the json-formatted log file generated from a nipype
        workflow execution

    Returns
    -------
    nodes_list : list
        a list of python dictionaries containing the runtime info
        for each nipype node
    """

    # Init variables
    with open(logfile, "r") as content:
        # read file separating each line
        lines = content.readlines()

    nodes_list = [json.loads(l) for l in lines]

    # Return list of nodes
    return nodes_list


def calculate_resource_timeseries(events, resource):
    """
    Given as event dictionary, calculate the resources used
    as a timeseries

    Parameters
    ----------
    events : dictionary
        a dictionary of event-based node dictionaries of the workflow
        execution statistics
    resource : string
        the resource of interest to return the time-series of;
        e.g. 'runtime_memory_gb', 'estimated_threads', etc

    Returns
    -------
    time_series : pandas Series
        a pandas Series object that contains timestamps as the indices
        and the resource amount as values
    """

    # Import packages
    import pandas as pd

    # Init variables
    res = OrderedDict()
    all_res = 0.0

    # Iterate through the events
    for _, event in sorted(events.items()):
        if event["event"] == "start":
            if resource in event and event[resource] != "Unknown":
                all_res += float(event[resource])
            current_time = event["start"]
        elif event["event"] == "finish":
            if resource in event and event[resource] != "Unknown":
                all_res -= float(event[resource])
            current_time = event["finish"]
        res[current_time] = all_res

    # Formulate the pandas timeseries
    time_series = pd.Series(data=list(res.values()), index=list(res.keys()))
    # Downsample where there is only value-diff
    ts_diff = time_series.diff()
    time_series = time_series[ts_diff != 0]

    # Return the new time series
    return time_series


def draw_lines(start, total_duration, minute_scale, scale):
    """
    Function to draw the minute line markers and timestamps

    Parameters
    ----------
    start : datetime.datetime obj
        start time for first minute line marker
    total_duration : float
        total duration of the workflow execution (in seconds)
    minute_scale : integer
        the scale, in minutes, at which to plot line markers for the
        gantt chart; for example, minute_scale=10 means there are lines
        drawn at every 10 minute interval from start to finish
    scale : integer
        scale factor in pixel spacing between minute line markers

    Returns
    -------
    result : string
        the html-formatted string for producing the minutes-based
        time line markers
    """

    # Init variables
    result = ""
    next_line = 220
    next_time = start
    num_lines = int(((total_duration // 60) // minute_scale) + 2)

    # Iterate through the lines and create html line markers string
    for line in range(num_lines):
        # Line object
        new_line = "<hr class='line' width='98%%' style='top:%dpx;'>" % next_line
        result += new_line
        # Time digits
        time = "<p class='time' style='top:%dpx;'> %02d:%02d </p>" % (
            next_line - 20,
            next_time.hour,
            next_time.minute,
        )
        result += time
        # Increment line spacing and digits
        next_line += minute_scale * scale
        next_time += datetime.timedelta(minutes=minute_scale)

    # Return html string for time line markers
    return result


def draw_nodes(start, nodes_list, cores, minute_scale, space_between_minutes, colors):
    """
    Function to return the html-string of the node drawings for the
    gantt chart

    Parameters
    ----------
    start : datetime.datetime obj
        start time for first node
    nodes_list : list
        a list of the node dictionaries
    cores : integer
        the number of cores given to the workflow via the 'n_procs'
        plugin arg
    total_duration : float
        total duration of the workflow execution (in seconds)
    minute_scale : integer
        the scale, in minutes, at which to plot line markers for the
        gantt chart; for example, minute_scale=10 means there are lines
        drawn at every 10 minute interval from start to finish
    space_between_minutes : integer
        scale factor in pixel spacing between minute line markers
    colors : list
        a list of colors to choose from when coloring the nodes in the
        gantt chart

    Returns
    -------
    result : string
        the html-formatted string for producing the minutes-based
        time line markers
    """

    # Init variables
    result = ""
    scale = space_between_minutes / minute_scale
    space_between_minutes = space_between_minutes / scale
    end_times = [
        datetime.datetime(
            start.year, start.month, start.day, start.hour, start.minute, start.second
        )
        for core in range(cores)
    ]

    # For each node in the pipeline
    for node in nodes_list:
        # Get start and finish times
        node_start = node["start"]
        node_finish = node["finish"]
        # Calculate an offset and scale duration
        offset = (
            (node_start - start).total_seconds() / 60
        ) * scale * space_between_minutes + 220
        # Scale duration
        scale_duration = (node["duration"] / 60) * scale * space_between_minutes
        if scale_duration < 5:
            scale_duration = 5
        scale_duration -= 2
        # Left
        left = 60
        for core in range(len(end_times)):
            if end_times[core] < node_start:
                left += core * 30
                end_times[core] = datetime.datetime(
                    node_finish.year,
                    node_finish.month,
                    node_finish.day,
                    node_finish.hour,
                    node_finish.minute,
                    node_finish.second,
                )
                break

        # Get color for node object
        color = random.choice(colors)
        if "error" in node:
            color = "red"

        # Setup dictionary for node html string insertion
        node_dict = {
            "left": left,
            "offset": offset,
            "scale_duration": scale_duration,
            "color": color,
            "node_name": node["name"],
            "node_dur": node["duration"] / 60.0,
            "node_start": node_start.strftime("%Y-%m-%d %H:%M:%S"),
            "node_finish": node_finish.strftime("%Y-%m-%d %H:%M:%S"),
        }
        # Create new node string
        new_node = (
            "<div class='node' style='left:%(left)spx;top:%(offset)spx;"
            "height:%(scale_duration)spx;background-color:%(color)s;'"
            "title='%(node_name)s\nduration:%(node_dur)s\n"
            "start:%(node_start)s\nend:%(node_finish)s'></div>" % node_dict
        )

        # Append to output result
        result += new_node

    # Return html string for nodes
    return result


def draw_resource_bar(
    start_time,
    finish_time,
    time_series,
    space_between_minutes,
    minute_scale,
    color,
    left,
    resource,
):
    """ """

    # Memory header
    result = "<p class='time' style='top:198px;left:%dpx;'>%s</p>" % (left, resource)
    # Image scaling factors
    scale = space_between_minutes / minute_scale
    space_between_minutes = space_between_minutes / scale

    # Iterate through time series
    ts_items = time_series.items()

    ts_len = len(time_series)
    for idx, (ts_start, amount) in enumerate(ts_items):
        if idx < ts_len - 1:
            ts_end = time_series.index[idx + 1]
        else:
            ts_end = finish_time
        # Calculate offset from start at top
        offset = (
            (ts_start - start_time).total_seconds() / 60.0
        ) * scale * space_between_minutes + 220
        # Scale duration
        duration_mins = (ts_end - ts_start).total_seconds() / 60.0
        height = duration_mins * scale * space_between_minutes
        if height < 5:
            height = 5
        height -= 2

        # Bar width is proportional to resource amount
        width = amount * 20

        if resource.lower() == "memory":
            label = "%.3f GB" % amount
        else:
            label = "%d threads" % amount

        # Setup dictionary for bar html string insertion
        bar_dict = {
            "color": color,
            "height": height,
            "width": width,
            "offset": offset,
            "left": left,
            "label": label,
            "duration": duration_mins,
            "start": ts_start.strftime("%Y-%m-%d %H:%M:%S"),
            "finish": ts_end.strftime("%Y-%m-%d %H:%M:%S"),
        }

        bar_html = (
            "<div class='bar' style='background-color:%(color)s;"
            "height:%(height).3fpx;width:%(width).3fpx;"
            "left:%(left)dpx; top:%(offset).3fpx;'"
            "title='%(label)s\nduration:%(duration).3f\n"
            "start:%(start)s\nend:%(finish)s'></div>"
        )
        # Add another bar to html line
        result += bar_html % bar_dict

    # Return bar-formatted html string
    return result


def generate_gantt_chart(
    logfile,
    cores,
    minute_scale=10,
    space_between_minutes=50,
    colors=["#7070FF", "#4E4EB2", "#2D2D66", "#9B9BFF"],
):
    """
    Generates a gantt chart in html showing the workflow execution based on a callback log file.
    This script was intended to be used with the MultiprocPlugin.
    The following code shows how to set up the workflow in order to generate the log file:

    Parameters
    ----------
    logfile : string
        filepath to the callback log file to plot the gantt chart of
    cores : integer
        the number of cores given to the workflow via the 'n_procs'
        plugin arg
    minute_scale : integer (optional); default=10
        the scale, in minutes, at which to plot line markers for the
        gantt chart; for example, minute_scale=10 means there are lines
        drawn at every 10 minute interval from start to finish
    space_between_minutes : integer (optional); default=50
        scale factor in pixel spacing between minute line markers
    colors : list (optional)
        a list of colors to choose from when coloring the nodes in the
        gantt chart


    Returns
    -------
    None
        the function does not return any value but writes out an html
        file in the same directory as the callback log path passed in

    Usage
    -----
    # import logging
    # import logging.handlers
    # from nipype.utils.profiler import log_nodes_cb

    # log_filename = 'callback.log'
    # logger = logging.getLogger('callback')
    # logger.setLevel(logging.DEBUG)
    # handler = logging.FileHandler(log_filename)
    # logger.addHandler(handler)

    # #create workflow
    # workflow = ...

    # workflow.run(plugin='MultiProc',
    #     plugin_args={'n_procs':8, 'memory':12, 'status_callback': log_nodes_cb})

    # generate_gantt_chart('callback.log', 8)
    """

    # add the html header
    html_string = """<!DOCTYPE html>
    <head>
        <style>
            #content{
                width:99%;
                height:100%;
                position:absolute;
            }

            .node{
                background-color:#7070FF;
                border-radius: 5px;
                position:absolute;
                width:20px;
                white-space:pre-wrap;
            }

            .line{
                position: absolute;
                color: #C2C2C2;
                opacity: 0.5;
                margin: 0px;
            }

            .time{
                position: absolute;
                font-size: 16px;
                color: #666666;
                margin: 0px;
            }

            .bar{
                position: absolute;
                height: 1px;
                opacity: 0.7;
            }

            .dot{
                position: absolute;
                width: 1px;
                height: 1px;
                background-color: red;
            }
            .label {
                width:20px;
                height:20px;
                opacity: 0.7;
                display: inline-block;
            }
        </style>
    </head>

    <body>
        <div id="content">
            <div style="display:inline-block;">
    """

    close_header = """
    </div>
    <div style="display:inline-block;margin-left:60px;vertical-align: top;">
        <p><span><div class="label" style="background-color:#90BBD7;"></div> Estimated Resource</span></p>
        <p><span><div class="label" style="background-color:#03969D;"></div> Actual Resource</span></p>
        <p><span><div class="label" style="background-color:#f00;"></div> Failed Node</span></p>
    </div>
    """

    # Read in json-log to get list of node dicts
    nodes_list = log_to_dict(logfile)

    # Create the header of the report with useful information
    start_node = nodes_list[0]
    last_node = nodes_list[-1]
    duration = (last_node["finish"] - start_node["start"]).total_seconds()

    # Get events based dictionary of node run stats
    events = create_event_dict(start_node["start"], nodes_list)

    # Summary strings of workflow at top
    html_string += (
        "<p>Start: " + start_node["start"].strftime("%Y-%m-%d %H:%M:%S") + "</p>"
    )
    html_string += (
        "<p>Finish: " + last_node["finish"].strftime("%Y-%m-%d %H:%M:%S") + "</p>"
    )
    html_string += "<p>Duration: " + "{0:.2f}".format(duration / 60) + " minutes</p>"
    html_string += "<p>Nodes: " + str(len(nodes_list)) + "</p>"
    html_string += "<p>Cores: " + str(cores) + "</p>"
    html_string += close_header
    # Draw nipype nodes Gantt chart and runtimes
    html_string += draw_lines(
        start_node["start"], duration, minute_scale, space_between_minutes
    )
    html_string += draw_nodes(
        start_node["start"],
        nodes_list,
        cores,
        minute_scale,
        space_between_minutes,
        colors,
    )

    # Get memory timeseries
    estimated_mem_ts = calculate_resource_timeseries(events, "estimated_memory_gb")
    runtime_mem_ts = calculate_resource_timeseries(events, "runtime_memory_gb")
    # Plot gantt chart
    resource_offset = 120 + 30 * cores
    html_string += draw_resource_bar(
        start_node["start"],
        last_node["finish"],
        estimated_mem_ts,
        space_between_minutes,
        minute_scale,
        "#90BBD7",
        resource_offset * 2 + 120,
        "Memory",
    )
    html_string += draw_resource_bar(
        start_node["start"],
        last_node["finish"],
        runtime_mem_ts,
        space_between_minutes,
        minute_scale,
        "#03969D",
        resource_offset * 2 + 120,
        "Memory",
    )

    # Get threads timeseries
    estimated_threads_ts = calculate_resource_timeseries(events, "estimated_threads")
    runtime_threads_ts = calculate_resource_timeseries(events, "runtime_threads")
    # Plot gantt chart
    html_string += draw_resource_bar(
        start_node["start"],
        last_node["finish"],
        estimated_threads_ts,
        space_between_minutes,
        minute_scale,
        "#90BBD7",
        resource_offset,
        "Threads",
    )
    html_string += draw_resource_bar(
        start_node["start"],
        last_node["finish"],
        runtime_threads_ts,
        space_between_minutes,
        minute_scale,
        "#03969D",
        resource_offset,
        "Threads",
    )

    # finish html
    html_string += """
        </div>
    </body>"""

    # save file
    with open(logfile + ".html", "w") as html_file:
        html_file.write(html_string)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..specialized import BRAINSMultiSTAPLE


def test_BRAINSMultiSTAPLE_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputCompositeT1Volume=dict(
            argstr="--inputCompositeT1Volume %s",
            extensions=None,
        ),
        inputLabelVolume=dict(
            argstr="--inputLabelVolume %s...",
        ),
        inputTransform=dict(
            argstr="--inputTransform %s...",
        ),
        labelForUndecidedPixels=dict(
            argstr="--labelForUndecidedPixels %d",
        ),
        outputConfusionMatrix=dict(
            argstr="--outputConfusionMatrix %s",
            hash_files=False,
        ),
        outputMultiSTAPLE=dict(
            argstr="--outputMultiSTAPLE %s",
            hash_files=False,
        ),
        resampledVolumePrefix=dict(
            argstr="--resampledVolumePrefix %s",
        ),
        skipResampling=dict(
            argstr="--skipResampling ",
        ),
    )
    inputs = BRAINSMultiSTAPLE.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSMultiSTAPLE_outputs():
    output_map = dict(
        outputConfusionMatrix=dict(
            extensions=None,
        ),
        outputMultiSTAPLE=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSMultiSTAPLE.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..svm import SVMTest


def test_SVMTest_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        classout=dict(
            argstr="-classout",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-testvol %s",
            extensions=None,
            mandatory=True,
        ),
        model=dict(
            argstr="-model %s",
            mandatory=True,
        ),
        multiclass=dict(
            argstr="-multiclass %s",
        ),
        nodetrend=dict(
            argstr="-nodetrend",
        ),
        nopredcensord=dict(
            argstr="-nopredcensord",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        options=dict(
            argstr="%s",
        ),
        out_file=dict(
            argstr="-predictions %s",
            extensions=None,
            name_template="%s_predictions",
        ),
        outputtype=dict(),
        testlabels=dict(
            argstr="-testlabels %s",
            extensions=None,
        ),
    )
    inputs = SVMTest.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SVMTest_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = SVMTest.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import ROIStats


def test_ROIStats_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        debug=dict(
            argstr="-debug",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        format1D=dict(
            argstr="-1Dformat",
            xor=["format1DR"],
        ),
        format1DR=dict(
            argstr="-1DRformat",
            xor=["format1D"],
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        mask=dict(
            argstr="-mask %s",
            deprecated="1.1.4",
            extensions=None,
            new_name="mask_file",
            position=3,
        ),
        mask_f2short=dict(
            argstr="-mask_f2short",
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        nobriklab=dict(
            argstr="-nobriklab",
        ),
        nomeanout=dict(
            argstr="-nomeanout",
        ),
        num_roi=dict(
            argstr="-numroi %s",
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            keep_extension=False,
            name_source="in_file",
            name_template="%s_roistat.1D",
            position=-1,
        ),
        quiet=dict(
            argstr="-quiet",
        ),
        roisel=dict(
            argstr="-roisel %s",
            extensions=None,
        ),
        stat=dict(
            argstr="%s...",
        ),
        zerofill=dict(
            argstr="-zerofill %s",
            requires=["num_roi"],
        ),
    )
    inputs = ROIStats.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ROIStats_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = ROIStats.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import MRDeGibbs


def test_MRDeGibbs_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        axes=dict(
            argstr="-axes %s",
            maxlen=2,
            minlen=2,
            sep=",",
            usedefault=True,
        ),
        bval_scale=dict(
            argstr="-bvalue_scaling %s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        grad_file=dict(
            argstr="-grad %s",
            extensions=None,
            xor=["grad_fsl"],
        ),
        grad_fsl=dict(
            argstr="-fslgrad %s %s",
            xor=["grad_file"],
        ),
        in_bval=dict(
            extensions=None,
        ),
        in_bvec=dict(
            argstr="-fslgrad %s %s",
            extensions=None,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        maxW=dict(
            argstr="-maxW %d",
            usedefault=True,
        ),
        minW=dict(
            argstr="-minW %d",
            usedefault=True,
        ),
        nshifts=dict(
            argstr="-nshifts %d",
            usedefault=True,
        ),
        nthreads=dict(
            argstr="-nthreads %d",
            nohash=True,
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_unr",
            position=-1,
        ),
    )
    inputs = MRDeGibbs.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MRDeGibbs_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = MRDeGibbs.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
